from model.lmia.common import *
from model.common.commonmodel import CommonModel, BuilderModel
from webform.lmiaportal.webformmodel import WebformModel
from model.common.finance import Finance
from model.common.contact import Contacts
from model.common.advertisement import (
    InterviewRecord,
    InterviewRecords,
    RecruitmentSummary,
)
import os
from model.common.person import Person


class Emp5626(BaseModel):
    named: bool
    is_in_seasonal_industry: bool
    start_month: Optional[str]
    end_month: Optional[str]
    last_canadian_number: Optional[int]
    last_tfw_number: Optional[int]
    current_canadian_number: Optional[int]
    current_tfw_number: Optional[int]
    tp_waivable: bool
    waive_creteria: Optional[str]
    has_finished_tp: Optional[bool]
    finished_tp_result: Optional[str]
    activity1_title: Optional[str]
    activity1_description: Optional[str]
    activity1_outcome: Optional[str]
    activity1_comment: Optional[str]
    activity2_title: Optional[str]
    activity2_description: Optional[str]
    activity2_outcome: Optional[str]
    activity2_comment: Optional[str]
    activity3_title: Optional[str]
    activity3_description: Optional[str]
    activity3_outcome: Optional[str]
    activity3_comment: Optional[str]
    activity4_title: Optional[str]
    activity4_description: Optional[str]
    activity4_outcome: Optional[str]
    activity4_comment: Optional[str]
    activity5_title: Optional[str]
    activity5_description: Optional[str]
    activity5_outcome: Optional[str]
    activity5_comment: Optional[str]

    @root_validator
    def checkAnswers(cls, values):
        questions = [
            "tp_waivable",
            "has_finished_tp",
        ]
        explanations = [
            "waive_creteria",
            "finished_tp_result",
        ]
        qas = dict(zip(questions, explanations))
        for k, v in qas.items():
            if values.get(k) and not values.get(v):
                raise ValueError(
                    f"Since {k} is true, but you did not answer the question {v} in info-position sheet"
                )
        is_in_seasonal_industry = values.get("is_in_seasonal_industry", None)
        is_seasonal_vars = [
            values.get(var, None)
            for var in [
                "start_month",
                "end_month",
                "last_canadian_number",
                "last_tfw_number",
                "current_canadian_number",
                "current_tfw_number",
            ]
        ]
        if is_in_seasonal_industry and not all(v is not None for v in is_seasonal_vars):
            raise ValueError(
                "The position is in seasonal industry, but you didn't input all the information required"
            )

        return values


class Personal(Person):
    last_name: Optional[str]
    first_name: Optional[str]
    sex: Optional[str]
    dob: Optional[date]
    citizen: Optional[str]


class M5626Model(BaseModel, BuilderModel):
    lmiacase: LmiaCase
    general: General
    lmi: Lmi
    emp5626: Emp5626
    eraddress: List[ErAddress]
    contact: List[Contact]
    finance: List[Finance]
    joboffer: Joboffer
    personal: Personal
    personalassess: PersonalAssess
    position: Position
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    rcic: Rcic
    recruitmentsummary: RecruitmentSummary

    def checkTFW(self):
        if self.emp5626.named and not all(
            [
                self.personal.first_name,
                self.personal.last_name,
                self.personal.sex,
                self.personal.citizen,
                self.personal.dob,
            ]
        ):
            raise ValueError(
                "Since TFW is name required, you haven't input all information required in info-personal sheet."
            )

    def make_web_form(self, output_json, upload_dir, rcic, *args, **kwargs):
        args = {
            "model_variable": "5626",
            "app": self,
            "output_json": output_json,
            "upload_dir": upload_dir,
            "rcic": rcic,
        }
        wf = WebformModel(**args)
        wf.save()

    def make_pdf_form(self):
        pass

    def context(self, *args, **kwargs):
        interview = InterviewRecords(self.interviewrecord)
        primary_contact = Contacts(self.contact).primary
        return {
            **self.dict(),
            "formatted_finance": [
                {
                    "year": f.year,
                    "formatted_revenue": f.formatted_revenue,
                    "formatted_net_income": f.formatted_net_income,
                    "formatted_retained_earning": f.formatted_retained_earning,
                }
                for f in self.finance
            ],
            "num_of_job_posts": Advertisements(self.advertisement).amount,
            "why_qualified": self.personalassess.why_qualified,
            "adv_summary": [
                {
                    "start_date": a.start_date,
                    "end_date": a.end_date,
                    "media": a.media,
                    "days": a.days,
                }
                for a in self.advertisement
            ],
            "high_wage": True
            if float(self.joboffer.hourly_rate) >= self.lmiacase.provincial_median_wage
            else False,
            "rs": {
                "resume_num": interview.resume_num,
                "canadian_num": interview.canadian_num,
                "unknown_num": interview.unknown_num,
                "foreigner_num": interview.foreigner_num,
                "total_canadian": interview.total_canadian,
                "total_interviewed_canadians": interview.total_interviewed_canadians,
                "canadian_records": interview.canadian_records,
            },
            "primary_contact": primary_contact,
            "within_2years": self.general.within_2years,
            "lmi_benefits": [
                self.lmi.job_creation_benefit,
                self.lmi.skill_transfer_benefit,
                self.lmi.fill_shortage_benefit,
                self.lmi.other_benefit,
            ],
            "duration": self.lmiacase.duration,
            "why_qualified": [
                item
                for item in [
                    self.personalassess.work_experience_brief,
                    self.personalassess.education_brief,
                    self.personalassess.competency_brief,
                    self.personalassess.language_brief,
                    self.personalassess.performance_remark,
                ]
                if item
            ],
            "employee_name": self.personal.full_name,
        }


class M5626ModelE(CommonModel, M5626Model):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/recruitment.xlsx",
            "excel/pa.xlsx",
            "excel/er.xlsx",
            "excel/rep.xlsx",
            "excel/lmia.xlsx",
        ]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from model.common.user import User
from fastapi import APIRouter, Depends, HTTPException, status
from bson import ObjectId
from apps.config import Default, db
import schemas
from apps.api.hashing import Hash
from apps.api import oauth2
from apps.api import schemas

user_collection = db.user
router = APIRouter(
    prefix="/user", tags=["Users"], responses={404: {"description": "Nof found"}}
)


def email_existed(email: str):
    return user_collection.find_one({"email": email})


def is_admin(current_user: User):
    if current_user.get("role") != "admin":
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="The operation needs admin access right",
        )
    return True


@router.get("/")
async def read_users(current_user: User = Depends(oauth2.get_current_user)):
    users = list(user_collection.find())
    users1 = schemas.users_schema(users)
    return {"status_code": status.HTTP_200_OK, "data": users1}


@router.get("/{id}")
async def retrieve(id: str, current_user: User = Depends(oauth2.get_current_user)):
    user = user_collection.find_one({"_id": ObjectId(id)})
    return {"status_code": status.HTTP_200_OK, "data": schemas.user_schema(user)}


@router.post("/")
async def create(user: User):
    if email_existed(user.email):
        raise HTTPException(
            status_code=status.HTTP_406_NOT_ACCEPTABLE,
            detail=f"{user.email} is existed in the system.",
        )
    user.password = Hash.hash(user.password)
    user.permission.extend(Default.user_permission)
    new_id = user_collection.insert_one(user.dict()).inserted_id
    new_user = user_collection.find_one({"_id": new_id})
    return {"status_code": status.HTTP_200_OK, "data": schemas.user_schema(new_user)}


@router.put("/{id}")
async def update(
    id: str, user: User, current_user: User = Depends(oauth2.get_current_user)
):
    if user == current_user or is_admin(current_user):
        user_collection.update_one({"_id": ObjectId(id)}, {"$set": user.dict()})
        updated_one = user_collection.find_one({"_id": ObjectId(id)})
        return {
            "status_code": status.HTTP_200_OK,
            "data": schemas.user_schema(updated_one),
        }


@router.put("/password/{id}")
async def password(
    id: str, password: str, current_user: User = Depends(oauth2.get_current_user)
):
    user = user_collection.find_one({"_id": ObjectId(id)})
    user["password"] = Hash.hash(password)
    if user["email"] == current_user["sub"] or is_admin(current_user):
        user_collection.update_one({"_id": ObjectId(id)}, {"$set": user})
        updated_one = user_collection.find_one({"_id": ObjectId(id)})
        return {
            "status_code": status.HTTP_200_OK,
            "data": schemas.user_schema(updated_one),
        }


@router.delete("/{id}")
async def delete(id: str, current_user: User = Depends(oauth2.get_current_user)):
    is_admin(current_user)
    try:
        res = user_collection.delete_one({"_id": ObjectId(id)})
        if res.deleted_count > 0:
            return {"status_code": status.HTTP_204_NO_CONTENT, "data": []}
        return {
            "status_code": status.HTTP_404_NOT_FOUND,
            "detail": "Delete unseccessfully.",
        }
    except Exception as e:
        return {
            "status_code": status.HTTP_404_NOT_FOUND,
            "detail": "Delete unseccessfully.",
        }


@router.post("/permission")
async def permission(
    request: schemas.Permission,
    current_user: User = Depends(oauth2.get_current_user),
):
    # print("In permission",type(current_user),current_user)
    is_admin(current_user)
    user = user_collection.find_one({"email": request.user_email})
    if not user:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"User {request.user_email} is not in the system",
        )

    # add permissions
    permissions = user["permission"]
    for permission in request.permissions:
        if request.remove:
            if permission in permissions:
                permissions.remove(permission)
        else:
            if permission not in permissions:
                permissions.append(permission)

    # update
    user_obj = User(**user)
    await update(user["_id"], user_obj)
    return {"permission": user.get("permission")}


@router.post("/role")
async def role(
    request: schemas.Role,
    current_user: User = Depends(oauth2.get_current_user),
):
    is_admin(current_user)
    user = user_collection.find_one({"email": request.user_email})
    if not user:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"User {request.user_email} is not in the system",
        )

    # set role
    user["role"] = request.role
    # update
    user_obj = User(**user)
    await update(user["_id"], user_obj)
    return {"role": user.get("role")}



import os, sys
from pathlib import Path

# Get project's home directory,
# BASEDIR=os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
BASEDIR = Path(__file__).parents[3]
# All data directory
DATADIR = BASEDIR / "data"
# Insert the BASEDIR to system path
sys.path.insert(0, os.fspath(BASEDIR))



from pydantic import BaseModel, validator
from datetime import date, timedelta
from typing import Optional, List
from model.common.mixins import DurationMixin
from model.common.utils import makeList, Duration
from dateutil.parser import parse, ParserError
from model.common.employmentbase import EmploymentBase


class EmploymentHistory:
    def __init__(self, employment_list: List[EmploymentBase]):
        self.employment_list = employment_list
        for emp in employment_list:
            if not isinstance(emp, EmploymentBase):
                raise ValueError(
                    f"{employment_list} is not a list of EmploymentBase objects"
                )

    @property
    def initial_start_date(self):
        return min([emp.start_date for emp in self.employment_list])

    @property
    def final_end_date(self):
        end_dates = []
        for emp in self.employment_list:
            if emp.is_present:
                return "Present"
            end_dates.append(emp.end_date)
        return max(end_dates) or "Present"

    @property
    def position_number(self):
        return len(set([emp.job_title for emp in self.employment_list]))

    @property
    def position_number_say(self):
        number = len(set([emp.job_title for emp in self.employment_list]))
        return str(number) + " position" if number <= 1 else str(number) + " positions"

    def qualified_employment(self, program):
        return [job for job in self.employment_list if getattr(job, program)]



from .context import DATADIR
from typing import List
from model.common.commonmodel import CommonModel, BuilderModel
from model.bcpnp.data import Contact, General, JobOffer, PersonalAssess, Bcpnp
from model.common.jobposition import PositionBase
from model.common.person import Person
from model.common.advertisement import (
    Advertisement,
    Advertisements,
    InterviewRecord,
    RecruitmentSummary,
    InterviewRecords,
)
from model.common.contact import Contacts
from model.common.wordmaker import WordMaker
import os
from pydantic import BaseModel


class Personal(Person):
    def __str__(self):
        return self.full_name


class Position(PositionBase):
    pass


class RecommendationLetterModel(BaseModel, BuilderModel):
    bcpnp: Bcpnp
    general: General
    contact: List[Contact]
    position: Position
    personal: Personal
    joboffer: JobOffer
    personalassess: PersonalAssess
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    recruitmentsummary: RecruitmentSummary

    @property
    def selected_contact(self):
        contacts = Contacts(self.contact)
        return contacts.preferredContact

    @property
    def summary(self):
        return InterviewRecords(self.interviewrecord)

    @property
    def advertisements(self):
        return Advertisements(self.advertisement)

    @property
    def person(self):
        return {
            "full_name": self.personal.full_name,
            "attributive": self.personal.attributive,
            "object": self.personal.object,
            "subject": self.personal.subject,
            "short_name": self.personal.short_name,
        }

    def context(self, *args, **kwargs):
        return {
            **self.dict(),
            "resume_num": self.summary.resume_num,
            "canadian_num": self.summary.canadian_num,
            "unknown_num": self.summary.unknown_num,
            "foreigner_num": self.summary.foreigner_num,
            "total_canadian": self.summary.total_canadian,
            "total_interviewed_canadians": self.summary.total_interviewed_canadians,
            "canadian_records": self.summary.canadian_records,
            "contact": self.selected_contact,
            "advertisement": self.advertisements,
            "personal": self.person,
            "work_start_date": self.joboffer.start_date_say,
            "joboffer_date": self.joboffer.date_of_offer,
        }

    def make_pdf_form(self, *args, **kwargs):
        pass

    def make_web_form(self, *args, **kwargs):
        pass


class RecommendationLetterModelE(CommonModel, RecommendationLetterModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/er.xlsx",
            "excel/pa.xlsx",
            "excel/recruitment.xlsx",
            "excel/bcpnp.xlsx",
            "excel/rep.xlsx",
        ]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from fastapi import APIRouter, Depends, HTTPException, status
from starlette.responses import FileResponse
from apps.config import db
from apps.api.show import show_exception
from apps.api.models import ImmMake, ImmCheck, ImmPdfform, ImmWebform, ImmWord, ImmRun
import os
from apps.api import oauth2
from model.common.user import User

case = db.case

router = APIRouter(
    prefix="/case", tags=["Cases"], responses={404: {"description": "Nof found"}}
)


def permission_check(command: str, user: User):
    user_permissions = user.get("permission")
    if not command in user_permissions and not "all" in user_permissions:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="You have no permission for make command.",
        )


@router.post("/make")
async def make(data: ImmMake, current_user: User = Depends(oauth2.get_current_user)):
    permission_check("make", current_user)
    try:
        filename = data.make()
        return FileResponse(filename)
    except Exception as e:
        show_exception(e)


@router.post("/check")
async def check(data: ImmCheck, current_user: User = Depends(oauth2.get_current_user)):
    permission_check("check", current_user)
    try:
        return data.check()
    except Exception as e:
        show_exception(e)


@router.post("/word")
async def word(data: ImmWord, current_user: User = Depends(oauth2.get_current_user)):
    permission_check("word", current_user)
    try:
        filename = data.word()
        return FileResponse(filename)
    except Exception as e:
        show_exception(e)


@router.post("/webform")
async def webform(
    data: ImmWebform, current_user: User = Depends(oauth2.get_current_user)
):
    permission_check("webform", current_user)
    try:
        filename = data.webform()
        return FileResponse(filename)
    except Exception as e:
        show_exception(e)


@router.post("/pdfform")
async def pdfform(
    data: ImmPdfform, current_user: User = Depends(oauth2.get_current_user)
):
    permission_check("pdfform", current_user)
    try:
        filename = data.pdfform()
        return FileResponse(filename)
    except Exception as e:
        show_exception(e)


@router.post("/run")
async def run(data: ImmRun, current_user: User = Depends(oauth2.get_current_user)):
    permission_check("run", current_user)
    try:
        return data.run()
    except Exception as e:
        show_exception(e)



import re, sys
from pydantic import BaseModel
from typing import Optional
from apps.models import get_models
from pathlib import Path
import uuid
from model.common.wordmaker import WordMaker
from apps.config import Default
from apps.api.context import BASEDIR


class ImmBase(BaseModel):
    model: str

    def getModel(self, model_type="original"):
        # get model, and its instance
        models = get_models(
            rcic_company_id_name=Default.rciccompany, temp_num=Default.temp_num
        )
        model = models.get(self.model)
        if not model:
            raise ValueError(f"{self.model} is not a valid model name.")
        class_list = model["class_list"]
        the_class = __import__(model["path"], fromlist=class_list)
        if model_type == "original":
            the_model = getattr(the_class, class_list[0])  # original model
        elif model_type == "excel":
            the_model = getattr(the_class, class_list[1])  # excel model
        else:
            the_model = getattr(the_class, class_list[2])  # future for database model
        return the_model


class ImmMake(ImmBase):
    def make(self):
        """Make excel based on model, return the file name"""
        # create excel
        the_model = self.getModel(model_type="excel")
        name = uuid.uuid4().hex + ".xlsx"
        excel_name = BASEDIR / "tmp" / name
        the_model(output_excel_file=excel_name)
        return excel_name


class ImmCheck(ImmBase):
    data: dict

    def check(self):
        the_model = self.getModel()
        model_obj = the_model(**self.data)
        if model_obj.dict() != {}:
            return {
                "success": "The model has been checked and everything seems good..."
            }


class ImmWord(ImmBase):
    data: dict
    output_name: str
    doctype: Optional[str]
    rciccompany: str = Default.rciccompany
    tempnum: int = Default.temp_num

    def word(self):
        the_model = self.getModel()
        models = get_models(
            rcic_company_id_name=self.rciccompany, temp_num=self.tempnum
        )
        template_docx_dict = models.get(self.model).get("docx_template")
        if not template_docx_dict:
            raise ValueError(
                f"There is no template {template_docx_dict} existing in model {self.model}"
            )

        # if there  is no args.document, which means the model has only one  default  dict element
        def getOnlyOneTemplate():
            the_only_one_temp = list(template_docx_dict.keys())
            temp_number = len(the_only_one_temp)
            if temp_number == 1:
                return template_docx_dict.get(the_only_one_temp[0])
            elif temp_number > 1:
                raise ValueError(
                    "Word templates more than one in the model, but you did not input the doc type abbreviation"
                )
            else:
                return None

        # if with doc type , get the specific template, or  without args.document, get the only one
        template_docx = (
            template_docx_dict.get(self.doctype)
            if self.doctype
            else getOnlyOneTemplate()
        )
        if not template_docx:
            raise ValueError(
                f"{self.doctype} is not a valid document type, please check the manual by typing:'imm man'to see valid word template for each model"
            )
        elif not Path(template_docx).exists():
            raise FileNotFoundError(
                f"{template_docx} is not existed. Please check your template base"
            )

        model_instance = the_model(**self.data)
        context = model_instance.context(doc_type=self.doctype)
        word_name = uuid.uuid4().hex + ".docx"
        word_name = BASEDIR / "tmp" / word_name
        wm = WordMaker(template_docx, context, word_name)
        wm.make()
        return word_name


class ImmWebform(ImmBase):
    data: dict
    output_name: str
    rcic: str = Default.rcic
    tempnum: int = Default.temp_num
    uploaddir: str = Default.uploaddir
    initial: bool = Default.initial
    previous: bool = Default.previous

    def webform(self):
        json_name = uuid.uuid4().hex + ".json"
        json_name = BASEDIR / "tmp" / json_name
        key_args = {
            "output_json": json_name,
            "upload_dir": self.uploaddir,
            "rcic": self.rcic,
            "initial": self.initial,
            "previous": self.previous,
        }
        the_model = self.getModel()
        app = the_model(**self.data)
        app.make_web_form(**key_args)
        return json_name


class ImmPdfform(ImmBase):
    data: dict
    output_name: str
    rcic: str = Default.rcic

    def pdfform(self):
        json_name = uuid.uuid4().hex + ".json"
        json_name = BASEDIR / "tmp" / json_name
        key_args = {"output_json": json_name, "rcic_id_name": self.rcic}
        the_model = self.getModel()
        app = the_model(**self.data)
        app.make_pdf_form(**key_args)
        return json_name


class ImmRun(ImmBase):
    data: dict

    def run(self):
        """Run app only"""
        the_model = self.getModel()
        model_obj = the_model(**self.data)
        result = ""
        if hasattr(model_obj, "result"):
            for k, v in model_obj.result.items():
                result += (f"{k}: {v}") + "\n"
            return {"success": result}
        else:
            raise ValueError(
                f"The model {self.model} has no excution function with return result."
            )



from .context import DATADIR
from model.common.commonmodel import CommonModel, BuilderModel
from model.bcpnp.data import JobOffer
from model.common.person import Person
from model.common.wordmaker import WordMaker
import os
from pydantic import BaseModel


class Personal(Person):
    def __str__(self):
        return self.full_name


class JobDescriptionModel(BaseModel, BuilderModel):
    personal: Personal
    joboffer: JobOffer

    def context(self, *args, **kwargs):
        context = {**self.dict(), "requirements": self.joboffer.requirements}
        return context

    def make_pdf_form(self, *args, **kwargs):
        pass

    def make_web_form(self, *args, **kwargs):
        pass


class JobDescriptionModelE(CommonModel, JobDescriptionModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/er.xlsx", "excel/pa.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from apps.api.routers.authentication import verify_token

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="login")


def get_current_user(data: str = Depends(oauth2_scheme)):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )

    return verify_token(data, credentials_exception)



from termcolor import colored
from functools import reduce
from typing import List
from model.common.address import Address
from model.common.phone import Phone
from model.common.trperson import (
    COR,
    PersonId,
    Personal,
    Marriage,
    Education,
    Employment,
)
import json

from model.common.tr import TrCaseIn, TrBackground, SpInCanada
from model.common.commonmodel import CommonModel, BuilderModel
from pdfform.tr.fb5709 import FormBuilder5709
from pydantic import BaseModel


class M5709Model(BaseModel, BuilderModel):
    personal: Personal
    marriage: Marriage
    personid: List[PersonId]
    address: List[Address]
    education: List[Education]
    employment: List[Employment]
    phone: List[Phone]
    cor: List[COR]
    trcasein: TrCaseIn
    spincanada: SpInCanada
    trbackground: TrBackground

    def make_pdf_form(self, output_json, *args, **kwargs):
        pf = FormBuilder5709(self)
        form = pf.get_form()
        with open(output_json, "w") as output:
            json.dump(form.actions, output, indent=3, default=str)
        print(colored(f"{output_json} has been created. ", "green"))

    def make_web_form(self, output_json, upload_dir, rcic, *args, **kwargs):
        raise ValueError("This model doesn't have webform...")

    def context(self, *args, **kwargs):
        raise ValueError("This model doesn't have webform...")


class M5709ModelE(CommonModel, M5709Model):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/tr.xlsx", "excel/pa.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from pdfform.form_controls_win import (
    Control,
    RadioButtonList,
    TextField,
    RadioButton,
    CheckBox,
    DropdownList,
    DateField,
    Skip,
    Pause,
    OutputInfo,
    Button,
)


class ApplicationForm:
    """Application for filling PDF form according to json file"""

    def __init__(self, actions=None, verbose=False):
        self.verbose = verbose
        if actions is not None:
            self._actions = list(actions)
        else:
            self._actions = []

    def add_step(self, action: dict):
        match action.get("action_type"):
            case "Skip":
                control = Skip(
                    action.get("times"), action.get("pause"), verbose=self.verbose
                )
                self._actions.append(control)
            case "TextField":
                control = TextField(
                    action.get("data"), pause=action.get("pause"), verbose=self.verbose
                )
                self._actions.append(control)
            case "RadioButton":
                control = RadioButton(
                    action.get("data"),
                    pause=action.get("pause"),
                    verbose=self.verbose,
                )
                self._actions.append(control)

            case "RadioButtonList":
                control = RadioButtonList(
                    pause=action.get("pause"),
                    verbose=self.verbose,
                    position=action.get("position"),
                )
                self._actions.append(control)

            case "CheckBox":
                control = CheckBox(
                    action.get("data"), pause=action.get("pause"), verbose=self.verbose
                )
                self._actions.append(control)
            case "DropdownList":
                control = DropdownList(
                    action.get("data"),
                    action.get("key"),
                    action.get("num"),
                    pause=action.get("pause"),
                    verbose=self.verbose,
                )
                self._actions.append(control)
            case "DateField":
                control = DateField(
                    action.get("date"),
                    noday=action.get("noday"),
                    pause=action.get("pause"),
                    verbose=self.verbose,
                )
                self._actions.append(control)
            case "Pause":
                control = Pause(action.get("pause"), verbose=self.verbose)
                self._actions.append(control)
            case "OutputInfo":
                control = OutputInfo(action.get("info"), verbose=self.verbose)
                self._actions.append(control)
            case "Button":
                control = Button(action.get("pause"), verbose=self.verbose)
                self._actions.append(control)
            case _:
                raise ValueError(f"{action.get('action_type')} is invalid")

    def fill_form(self, start: int = 0, verbose: bool = False):
        """Perform the fill pdf form action"""
        if start != 0:
            action = Skip(times=start, verbose=self.verbose)
            action.fill()
        step = 0
        for action in self._actions[start:]:
            if verbose:
                print(
                    f"step: {step + start}---skip: {step+start+Skip.addtional_skip-3}",
                    end="\t",
                )
                step += 1
            action.fill()



from webform.lmiaportal.hourspay import HoursPay
from webform.lmiaportal.login import Login
from webform.lmiaportal.dashboard import Dashboard
from webform.lmiaportal.contact import Contact
from webform.lmiaportal.wage import Wage
from webform.lmiaportal.foreignnational import ForeignNational
from webform.lmiaportal.ee import EE
from webform.lmiaportal.hws_lws import HWS_LWS
from webform.lmiaportal.transationplan import TransitionPlan
from webform.lmiaportal.cap import Cap
from webform.lmiaportal.accommodation import Accommodation
from webform.lmiaportal.hourspay import HoursPay
from webform.lmiaportal.joboffer import JobOffer, JobOffer5593
from webform.lmiaportal.empbenefits import EmpBenefits
from webform.lmiaportal.recruitment import Recruitment
from webform.lmiaportal.lmbenefits import LmBenefits
from webform.lmiaportal.layoff import Layoff
from webform.lmiaportal.upload import Upload
import json
from termcolor import colored
import os, dotenv

BASEDIR = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))


class WebformModel:
    """Generate json file for webform filling based on model

    Args:
        output_json:str,
        upload_dir:str,
        rcic: str = "jacky",
    """

    def __init__(self, model_variable, app, output_json, upload_dir, rcic="jacky"):
        self.model_variable = model_variable
        # self.source_excel = source_excel
        self.output_json = output_json
        self.upload_dir = upload_dir
        self.rcic = rcic
        self.app = app

    # models = {"5593": M5593Model, "5626": M5626Model, "5627": M5627Model}

    def save(self):
        # self.app = self.getApplicant()

        # login
        actions = self.login()
        # pages to fill
        actions += self.pages()

        # upload files
        actions += self.upload()
        # output actions
        # print(json.dumps(actions, indent=3, default=str))

        # output json
        self.output(actions)

    def getApplicant(self):
        # get lmia data
        the_model = WebformModel.models.get(self.model_variable)
        if not the_model:
            print(
                colored(
                    f"The model {self.model_variable} is not existed, please check.",
                    "red",
                )
            )
            return None

        app = the_model(excels=[self.source_excel])
        if not app:
            print(
                colored(
                    f"The model {self.source_excel} is not existed, please check.",
                    "red",
                )
            )
            return None
        return app

    def login(self):
        path = os.path.abspath(os.path.join(os.path.expanduser("~"), ".immenv"))
        config = dotenv.dotenv_values(path)
        rcic_account = {
            "account": config.get(self.rcic + "_lmiaportal_account"),
            "password": config.get(self.rcic + "_lmiaportal_password"),
            "security_answers": config.get(self.rcic + "_lmiaportal_sequrity_answers"),
        }
        if (
            not rcic_account["account"]
            or not rcic_account["password"]
            or not rcic_account["security_answers"]
        ):
            print(
                colored(
                    f"{self.rcic}'s prportal account, password, and/or security answers is not existed. Check the .immenv file in your home directory and add your profile",
                    "red",
                )
            )
            exit(1)

        security_dict = json.loads(rcic_account["security_answers"])
        l = Login(
            rcic_account["account"],
            rcic_account["password"],
            security_dict,
        )
        return l.actions

    def pages(self):
        actions = []
        modules = {
            "5593": [
                Dashboard,
                Contact,
                Wage,
                ForeignNational,
                EE,
                HoursPay,
                JobOffer5593,
                EmpBenefits,
                Recruitment,
                LmBenefits,
                Layoff,
            ],
            "5626": [
                Dashboard,
                Contact,
                HWS_LWS,
                TransitionPlan,
                HoursPay,
                JobOffer,
                EmpBenefits,
                Recruitment,
                LmBenefits,
                Layoff,
            ],
            "5627": [
                Dashboard,
                Contact,
                HWS_LWS,
                Accommodation,
                Cap,
                HoursPay,
                JobOffer,
                EmpBenefits,
                Recruitment,
                LmBenefits,
                Layoff,
            ],
        }
        if self.model_variable not in modules:
            raise ValueError(
                f"There is no module named: {self.model_variable} in webformmodel.py"
            )

        for module in modules.get(self.model_variable):
            actions += module(self.app).actions

        return actions

    def upload(self):
        actions = []
        if not os.path.exists(self.upload_dir):
            raise ValueError(f"The directory {self.upload_dir} is not existed")

        upload = Upload(self.app, self.upload_dir)
        actions += upload.actions
        return actions

    def output(self, actions):
        if os.path.exists(self.output_json):
            overwrite = input(
                colored(f"The file {self.output_json} is existed, overwrite?", "red")
            )
            if overwrite and overwrite[0].lower() != "y":
                return
        with open(self.output_json, "w") as f:
            json.dump(actions, f, indent=3, default=str)
            print(colored(f"{self.output_json} has been saved", "green"))



from asyncore import read
from datetime import date, datetime
from pydantic import BaseModel
from typing import Union, Optional


class Language(BaseModel):
    """Base class for IELTS, CELPIP, TEF AND TCF"""

    reading: Union[int, float]
    writting: Union[int, float]
    listening: Union[int, float]
    speaking: Union[int, float]
    test_date: Optional[date]
    sign_date: Optional[date]

    def is_valid(self, the_day=date.today()):
        if not self.sign_date:
            raise ValueError("There is no sign data in Language object")
        if not isinstance(the_day, date):
            the_day = datetime.strptime(the_day, "%Y-%m-%d").date()
        return True if (the_day - self.sign_date).days < 730 else False

    def __str__(self):
        return f"{self.reading},{self.writting},{self.listening},{self.speaking}"

    class Meta:
        abstract = True


class CLB:
    """CLB to IELTS., CELPIP, TEF and TCF converter"""

    # 1. IELTS->CLB level, Reading	Writing	Listening	Speaking
    ielts_table = [
        [10, 8.0, 7.5, 8.5, 7.5],
        [9, 7.0, 7.0, 8.0, 7.0],
        [8, 6.5, 6.5, 7.5, 6.5],
        [7, 6.0, 6.0, 6.0, 6.0],
        [6, 5.0, 5.5, 5.5, 5.5],
        [5, 4.0, 5.0, 5.0, 5.0],
        [4, 3.5, 4.0, 4.5, 4.0],
    ]
    tef_table = [
        [10, 263, 277, 393, 415, 316, 333, 393, 415],
        [9, 248, 262, 371, 392, 298, 315, 371, 392],
        [8, 233, 247, 349, 370, 280, 297, 349, 370],
        [7, 207, 232, 310, 348, 249, 279, 310, 348],
        [6, 181, 206, 271, 309, 217, 248, 271, 309],
        [5, 151, 180, 226, 270, 181, 216, 226, 270],
        [4, 121, 150, 181, 225, 145, 180, 181, 225],
    ]
    tcf_table = [
        [10, 549, 699, 16, 20, 549, 699, 16, 20],
        [9, 524, 548, 14, 15, 523, 548, 14, 15],
        [8, 499, 523, 12, 13, 503, 522, 12, 13],
        [7, 453, 498, 10, 11, 458, 502, 10, 11],
        [6, 406, 452, 7, 9, 398, 457, 7, 9],
        [5, 375, 405, 6, 6, 369, 397, 6, 6],
        [4, 342, 374, 4, 5, 331, 368, 4, 5],
    ]

    def __init__(self, clb_level):
        self.level = clb_level
        self.level = 10 if self.level > 10 else self.level

    # input CLB level, return r w l s of IELTS
    def to_ielts(self):
        if self.level < 4:
            return [0, 0, 0, 0]
        index = 10 - self.level
        return CLB.ielts_table[index][1:5]

    # 输入CLB值 返回对应的tef r w l s分数列表
    def to_tef(self):
        if self.level < 4:
            return [0, 0, 0, 0, 0, 0, 0, 0]
        index = 10 - self.level
        return CLB.tef_table[index][1:9]

        # 输入CLB值 返回对应的tef r w l s分数列表

    def to_tcf(self):
        if self.level < 4:
            return [0, 0, 0, 0, 0, 0, 0, 0]
        index = 10 - self.level
        return CLB.tcf_table[index][1:9]


class IELTS(Language):
    reading: float
    writting: float
    listening: float
    speaking: float
    type_of_test: Optional[str] = "General"
    report_number: Optional[str]

    # def __init__(self, ielts: dict):
    #     super().__init__(**ielts)
    #     self.type_of_test = ielts.get("type_of_test") or "General"
    #     self.report_number = ielts.get("report_number") or ""

    def _ielts_2_clbs(self):
        ielts = [self.reading, self.writting, self.listening, self.speaking]
        clb_level = []
        for column in range(1, 5):  # 从读到说，4个列，从1开始，开始找对应的clb级别
            factor = float(ielts[column - 1])  # r w l s每个子项
            for level in range(0, 7):  # 从CLB 级别10一直找到4 7行
                if factor > CLB.ielts_table[0][column]:  # 如果语言成绩高于最高分，按照最高分算
                    clb_level.append(10)
                    break
                if factor < CLB.ielts_table[6][column]:  # 如果低于最低分，则CLB Level =0 跳出本列循环
                    clb_level.append(0)
                    break
                # 要算每个子项的CLB值
                if CLB.ielts_table[level][column] == ielts[column - 1]:
                    clb_level.append(CLB.ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                    break
            # 如果某项分数不在阵列中
            for level in range(1, 7):
                if (
                    CLB.ielts_table[level][column]
                    < factor
                    < CLB.ielts_table[level - 1][column]
                ):
                    clb_level.append(CLB.ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                    break
        return clb_level

    # public class mothod for returning single factor's clb level. reading:0, writing:1, listening:2, speaking:3
    @classmethod
    def ielts_2_clb(self, score, factor_index):
        column = factor_index + 1
        for level in range(0, 7):  # 从CLB 级别10一直找到4 7行
            if score > CLB.ielts_table[0][column]:  # 如果语言成绩高于最高分，按照最高分算
                return 10
            if score < CLB.ielts_table[6][column]:  # 如果低于最低分，则CLB Level =0 跳出本列循环
                return 0
            # 要算每个子项的CLB值
            if CLB.ielts_table[level][column] == score:
                return CLB.ielts_table[level][0]  # 得到该雅思成绩对应的CLB 级别
        # 如果某项分数不在阵列中
        for level in range(1, 7):
            if (
                CLB.ielts_table[level][column]
                < score
                < CLB.ielts_table[level - 1][column]
            ):
                return CLB.ielts_table[level][0]  # 得到该雅思成绩对应的CLB 级别

    @property
    def clb_r(self):
        return self._ielts_2_clbs()[0]

    @property
    def clb_w(self):
        return self._ielts_2_clbs()[1]

    @property
    def clb_l(self):
        return self._ielts_2_clbs()[2]

    @property
    def clb_s(self):
        return self._ielts_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r, self.clb_w, self.clb_l, self.clb_s)


class CELPIP(Language):
    reading: int
    writting: int
    listening: int
    speaking: int
    registration_number: Optional[str]
    pin_number: Optional[str]

    # def __init__(self, celpip: dict):
    #     super().__init__(**celpip)
    #     self.registration_number = celpip.get("registration_number") or ""
    #     self.pin_number = celpip.get("pin_number") or ""

    @property
    def clb_r(self):
        return self.reading

    @property
    def clb_w(self):
        return self.writting

    @property
    def clb_l(self):
        return self.listening

    @property
    def clb_s(self):
        return self.speaking

    @property
    def clb(self):
        return min(self.reading, self.writting, self.listening, self.speaking)


class TEF(Language):
    # def __init__(self, tef: dict):
    #     super().__init__(**tef)

    def _tef_2_clbs(self):
        if self.reading not in range(121, 278):
            raise ValueError("TEF reading score is invalid")
        if self.writting not in range(181, 416):
            raise ValueError("TEF writing score is invalid")
        if self.listening not in range(145, 334):
            raise ValueError("TEF listening score is invalid")
        if self.speaking not in range(181, 416):
            raise ValueError("TEF speaking score is invalid")

        rwls = [self.reading, self.writting, self.listening, self.speaking]
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(
                    CLB.tef_table[row][start], CLB.tef_table[row][end] + 1
                ):
                    clb.append(CLB.tef_table[row][0])
                    break
        return clb

    @property
    def clb_r(self):
        return self._tef_2_clbs()[0]

    @property
    def clb_w(self):
        return self._tef_2_clbs()[1]

    @property
    def clb_l(self):
        return self._tef_2_clbs()[2]

    @property
    def clb_s(self):
        return self._tef_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r, self.clb_w, self.clb_l, self.clb_s)


class TCF(Language):
    # def __init__(self, tcf: dict):
    #     super().__init__(**tcf)

    def _tcf_2_clbs(self):
        if self.reading not in range(342, 700):
            raise ValueError("TCF reading score is invalid")
        if self.writting not in range(4, 21):
            raise ValueError("TCF writing score is invalid")
        if self.listening not in range(331, 700):
            raise ValueError("TCF listening score is invalid")
        if self.speaking not in range(4, 21):
            raise ValueError("TCF speaking score is invalid")

        rwls = [self.reading, self.writting, self.listening, self.speaking]
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(
                    CLB.tcf_table[row][start], CLB.tcf_table[row][end] + 1
                ):
                    clb.append(CLB.tcf_table[row][0])
                    break
        return clb

    @property
    def clb_r(self):
        return self._tcf_2_clbs()[0]

    @property
    def clb_w(self):
        return self._tcf_2_clbs()[1]

    @property
    def clb_l(self):
        return self._tcf_2_clbs()[2]

    @property
    def clb_s(self):
        return self._tcf_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r, self.clb_w, self.clb_l, self.clb_s)



from pydantic import ValidationError
from fastapi import HTTPException


def print_validation_error(e: ValidationError):
    error_msg = ""
    for err in e.errors():
        loc = "->".join(err["loc"]) + ": "
        msg = err["msg"]
        error_msg += loc + msg + "\n"
    return error_msg


def show_exception(e: Exception):
    if type(e) == ValidationError:
        raise HTTPException(status_code=422, detail=print_validation_error(e))
    else:
        raise HTTPException(status_code=422, detail=e)



from model.common.commonmodel import FormBuilderBase
from pdfform.jsonmaker import JsonMaker
from model.common.rcic import Rcics
from model.common.contact import Contacts
from datetime import date


class FormBuilderBcRep(FormBuilderBase):
    def __init__(self, data_bcrep: object, rcic_id_name: str, *arg, **kwargs):
        self.data_bcrep = data_bcrep
        self.form = JsonMaker()
        self.rcic_id_name = rcic_id_name
        self.rcic = self.get_rcic(rcic_id_name)
        self.text_speed = 0.01
        self.skip_speed = 0.01

    def get_rcic(self, rcic_id_name):
        return Rcics(self.data_bcrep.rciclist).getRcicByIdName(rcic_id_name)

    def start(self):
        self.form.add_skip(1, pause=self.skip_speed)

    def add_applicant(self):
        self.form.add_text(self.data_bcrep.personal.last_name, pause=self.text_speed)
        self.form.add_text(self.data_bcrep.personal.first_name, pause=self.text_speed)
        self.form.add_date(self.data_bcrep.personal.dob, pause=self.text_speed)

    def add_employer(self):

        contact = Contacts(self.data_bcrep.contact).preferredContact
        # self.form.add_date(self.data_bcrep.general.legal_name, pause=self.text_speed)
        self.form.add_date("unigenius", pause=self.text_speed)
        self.form.add_text(contact.last_name, pause=self.text_speed)
        self.form.add_text(contact.first_name, pause=self.text_speed)
        self.form.add_skip(1, pause=self.skip_speed)

    @property
    def line1(self):
        # l1 = self.rcic.po_box + " " if self.rcic.po_box else ""
        l1 = self.rcic.unit + " " if self.rcic.unit else ""
        l1 += self.rcic.street_number + " " if self.rcic.street_number else ""
        l1 += self.rcic.street_name
        # l1 += ", " + self.rcic.district + ", " if self.rcic.district else ""
        return l1

    def add_rep(self):
        if not self.rcic:
            raise ValueError(f"RCIC with id_name {self.rcic_id_name} is not existed.")
        self.form.add_text(self.rcic.last_name, pause=self.text_speed)
        self.form.add_text(self.rcic.first_name, pause=self.text_speed)
        # employer name
        self.form.add_text(self.rcic.employer_legal_name, pause=self.text_speed)
        self.form.add_text(self.rcic.phone, pause=self.text_speed)
        self.form.add_skip(1, pause=self.skip_speed)
        self.form.add_text(self.rcic.rcic_number, pause=self.text_speed)
        self.form.add_text(self.rcic.email, pause=self.text_speed)
        self.form.add_text(self.line1, pause=self.text_speed)
        self.form.add_text(self.rcic.city, pause=self.text_speed)
        self.form.add_text(self.rcic.province, pause=self.text_speed)
        self.form.add_text(self.rcic.country, pause=self.text_speed)
        self.form.add_text(self.rcic.post_code, pause=self.text_speed)

        # check iccrc button
        self.form.add_checkbox(True)
        # skip one step for id number
        self.form.add_skip(1, pause=self.skip_speed)
        self.form.add_text(self.rcic.rcic_number, pause=self.text_speed)

    def add_date(self):
        # skip to rep date
        self.form.add_skip(5, pause=self.skip_speed)
        self.form.add_text(date.today().strftime("%d-%b-%Y"), pause=self.text_speed)
        self.form.add_text(date.today().strftime("%d-%b-%Y"), pause=self.text_speed)

    def get_form(self):
        self.start()
        self.add_applicant()
        self.add_employer()
        self.add_rep()
        self.add_date()
        return self.form



from pathlib import Path
from .local.config import Config

current_path = Path(__file__)
DATADIR = current_path.parents[1] / "data"

title = ["Command", "Arguments", "Options", "Remark"]

make = [
    "make",
    "model, excel_name",
    "",
    "Make excel based on the model and this excel name, with or without '.xlsx'",
]
check = [
    "check",
    "model, excel_name",
    "",
    "Check this excel data based on the model, with or without '.xlsx'",
]
run = [
    "run",
    "model, excel_name",
    "",
    "run a model for calculaton or something else with result return",
]


def word(options, remark):
    return [
        "word",
        "model, excel_name,[word_name],[doctype]",
        options,
        remark,
    ]


def webform(options):
    return [
        "webform",
        "model, excel_name, [outputjson]",
        options,
        "Make json file for filling web form",
    ]


def pdfform(options, remark):
    return [
        "pdfform",
        "model, excel_name, [outputjson]",
        options,
        remark,
    ]


helps = {
    "5593": [
        title,
        make,
        check,
        webform("--rcic,--uploaddir"),
        word(
            "",
            "Make word for rs:Recruitment Summary, et:Employer Training, sl: Submission Letter",
        ),
    ],
    "5626": [
        title,
        make,
        check,
        webform("--rcic,--uploaddir"),
        word(
            "",
            "Make word for rs:Recruitment Summary, et:Employer Training, sl: Submission Letter",
        ),
    ],
    "5627": [
        title,
        make,
        check,
        webform("--rcic,--uploaddir"),
        word(
            "",
            "Make word for rs:Recruitment Summary, et:Employer Training, sl: Submission Letter",
        ),
    ],
    "lmia-cap": [title, make, check, run],
    "exp": [
        title,
        make,
        check,
        word(
            "--tempnum",
            "Make word. doctype:  rs: resume, ec: employment certificate",
        ),
    ],
    "exp-rs": [
        title,
        make,
        check,
        word("--tempnum", "Make resume word"),
    ],
    "recruit-ja": [
        title,
        make,
        check,
        word("--rciccompany, --tempnum", "Make resume word"),
    ],
    "recruit-jo": [title, make, check, word("--tempnum", "Make resume word")],
    "recruit-rs": [title, make, check, word("", "Make recruitment summary doc")],
    "bcpnp-ert": [title, make, check, word("", "Make Employer Training doc")],
    "bcpnp-eet": [title, make, check, word("", "Make Employee Training doc")],
    "bcpnp-jd": [title, make, check, word("", "Make Job Description doc")],
    "bcpnp-rl": [
        title,
        make,
        check,
        word("", "Make Employer Recommendation Letter doc"),
    ],
    "bcpnp-edf": [
        title,
        make,
        check,
        pdfform("", "Make Employer Declaration Form json"),
    ],
    "bcpnp-rpf": [
        title,
        make,
        check,
        pdfform("--rcic", "Make Employer Declaration Form json"),
    ],
    "5476": [
        title,
        make,
        check,
        pdfform("--rcic", "Make 5476 Form json"),
    ],
}


def get_models(
    rcic_company_id_name: str = Config.rcic_company_id_name,
    temp_num: int = Config.default_template_number,
):
    models = {
        # Experience for resume and employment certificate
        "exp": {
            "path": "model.experience.experience",
            "class_list": ["ExperienceModel","ExperienceModelE"],
            "remark": "Experience module for Employment Certificate model",
            "docx_template": {
                "ec": DATADIR / "word" / "employment_certificate.docx",
                "rs": DATADIR / "word" / f"resume-regular{temp_num}.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate resume and employment certificate docx ",
                "helps": helps["exp"],
            },
        },
        "exp-rs": {
            "path": "model.experience.resume",
            "class_list": ["ResumeModel","ResumeModelE"],
            "remark": "Experience module for Resume model",
            "docx_template": {
                "rs": DATADIR / "word" / f"resume-regular{temp_num}.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate resume and employment certificate docx ",
                "helps": helps["exp-rs"],
            },
        },
        # Recruitment
        "recruit-ja": {
            "path": "model.recruit.jobad",
            "class_list": ["JobadModel","JobadModelE"],
            "remark": "Recruit module for Job Advertisement model",
            "docx_template": {
                "ja": DATADIR / "word" / "jobad.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate job advertisement docx ",
                "helps": helps["recruit-ja"],
            },
        },
        "recruit-jo": {
            "path": "model.recruit.joboffer",
            "class_list": ["JobofferModel","JobofferModelE"],
            "remark": "Recruitment module for Job Offer model",
            "docx_template": {
                "jo": DATADIR / "word" / f"joboffer{temp_num}.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate job offer docx ",
                "helps": helps["recruit-jo"],
            },
        },
        "recruit-rs": {
            "path": "model.recruit.recruitmentsummary",
            "class_list": ["RecruitmnetSummaryModel","RecruitmnetSummaryModelE"],
            "remark": "Recruitment module for Recruitment Summary model",
            "docx_template": {
                "ja": DATADIR / "word" / "lmia-rs.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate LMIA recruitment summary docx ",
                "helps": helps["recruit-rs"],
            },
        },
        # LMIA
        "lmia-st1": {
            "path": "model.lmia.stage1",
            "class_list": ["LmiaAssess","LmiaAssessE"],
            "remark": "LMIA module for stage 1 model (assessment)",
        },
        "lmia-st2": {
            "path": "model.lmia.stage2",
            "class_list": ["LmiaRecruitment","LmiaRecruitmentE"],
            "remark": "LMIA module for stage 2 model (recruitment)",
        },
        "lmia-st3": {
            "path": "model.lmia.stage3",
            "class_list": ["LmiaApplication","LmiaApplicationE"],
            "remark": "LMIA module for for stage 3 model (application)",
        },
        "lmia-rcic": {
            "path": "model.lmia.rcic",
            "class_list": ["LmiaRcic","LmiaRcicE"],
            "remark": "LMIA module for RCIC model (Plannning)",
        },
        "lmia-5593": {
            "path": "model.lmia.m5593",
            "class_list": ["M5593Model","M5593ModelE"],
            "docx_template": {
                "rs": DATADIR / "word" / "lmia-rs.docx",
                "et": DATADIR / "word" / "5593-et.docx",
                "sl": DATADIR / "word" / f"5593-sl-{rcic_company_id_name}.docx",
            },
            "remark": "LMIA module for EE doc generation application",
            "web_function": "✅ ",
            "help": {
                "description": "This model can automatically make docx, and make json file for filling web form ",
                "helps": helps["5593"],
            },
        },
        "lmia-5626": {
            "path": "model.lmia.m5626",
            "class_list": ["M5626Model","M5626ModelE"],
            "docx_template": {
                "rs": DATADIR / "word" / "lmia-rs.docx",
                "et": DATADIR / "word" / "5626-et.docx",
                "sl": DATADIR / "word" / f"5626-sl-{rcic_company_id_name}.docx",
            },
            "web_function": "✅ ",
            "remark": "LMIA module for HWS application",
            "help": {
                "description": "This model can automatically make docx, and make json file for filling web form ",
                "helps": helps["5626"],
            },
        },
        "lmia-5627": {
            "path": "model.lmia.m5627",
            "class_list": ["M5627Model","M5627ModelE"],
            "docx_template": {
                "rs": DATADIR / "word" / "lmia-rs.docx",
                "et": DATADIR / "word" / "5627-et.docx",
                "sl": DATADIR / "word" / f"5627-sl-{rcic_company_id_name}.docx",
            },
            "web_function": "✅ ",
            "remark": "LMIA module for LWS application",
            "help": {
                "description": "This model can automatically make docx, and make json file for filling web form ",
                "helps": helps["5627"],
            },
        },
        "lmia-cap": {
            "path": "model.lmia.cap",
            "class_list": ["CapModel","CapModelE"],
            "remark": "LMIA CAP module for LWS application",
            "help": {
                "description": "This model makes and checks excel, and calculate CAP ",
                "helps": helps["lmia-cap"],
            },
        },
        # BCPNP
        "bcpnp-ert": {
            "path": "model.bcpnp.employertraining",
            "class_list": ["EmployerTrainingModel","EmployerTrainingModelE"],
            "remark": "BCPNP module for Employer Training model",
            "docx_template": {
                "bcpnp-ert": DATADIR / "word" / "bcpnp-ert.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP employer training docx ",
                "helps": helps["bcpnp-ert"],
            },
        },
        "bcpnp-eet": {
            "path": "model.bcpnp.employeetraining",
            "class_list": ["EmployeeTrainingModel","EmployeeTrainingModelE"],
            "remark": "BCPNP module for Employee Training model",
            "docx_template": {
                "bcpnp-ert": DATADIR / "word" / "bcpnp-eet.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP employee training docx ",
                "helps": helps["bcpnp-eet"],
            },
        },
        "bcpnp-jd": {
            "path": "model.bcpnp.jobdescription",
            "class_list": ["JobDescriptionModel","JobDescriptionModelE"],
            "remark": "BCPNP module for Job Description model",
            "docx_template": {
                "bcpnp-jd": DATADIR / "word" / "bcpnp-jd.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP Job Description docx ",
                "helps": helps["bcpnp-jd"],
            },
        },
        "bcpnp-edf": {
            "path": "model.bcpnp.employerdeclaraton",
            "class_list": ["EmployerDeclaratonFormModel","EmployerDeclaratonFormModelE"],
            "remark": "BCPNP module for Employer Declaraton Form Model Form model",
            "pdf_function": "✅",
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP Employer Declaration Form data(Job Offer Form) ",
                "helps": helps["bcpnp-edf"],
            },
        },
        "bcpnp-rl": {
            "path": "model.bcpnp.recommendationletter",
            "class_list": ["RecommendationLetterModel","RecommendationLetterModelE"],
            "remark": "BCPNP module for Recommendation Letter model",
            "docx_template": {
                "bcpnp-rl": DATADIR / "word" / "bcpnp-rl.docx",
            },
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP Employer Recommendation Letter docx ",
                "helps": helps["bcpnp-rl"],
            },
        },
        "bcpnp-rpf": {
            "path": "model.bcpnp.mrep",
            "class_list": ["MRepModel","MRepModelE"],
            "remark": "BCPNP module for Representative Form model",
            "pdf_function": "✅",
            "help": {
                "description": "This model can make and check excel model, and generate BCPNP Rep form ",
                "helps": helps["bcpnp-rpf"],
            },
        },
        "bcpnp-pro": {
            "path": "webform.bcpnp.bcpnpmodel_pro",
            "class_list": ["BcpnpModelPro","BcpnpModelProE"],
            "remark": "BCPNP module for BCPNP Profile model",
            "web_function": "✅",
        },
        "bcpnp-reg": {
            "path": "webform.bcpnp.bcpnpmodel_reg",
            "class_list": ["BcpnpModelReg","BcpnpModelRegE"],
            "remark": "BCPNP module for BCPNP Registration model",
            "web_function": "✅",
        },
        "bcpnp-reg-ee": {
            "path": "webform.bcpnp.bcpnpmodel_reg",
            "class_list": ["BcpnpEEModelReg","BcpnpEEModelRegE"],
            "remark": "BCPNP module for BCPNP Registration model(EE)",
            "web_function": "✅",
        },
        "bcpnp-app": {
            "path": "webform.bcpnp.bcpnpmodel_app",
            "class_list": ["BcpnpModelApp","BcpnpModelAppE"],
            "remark": "BCPNP module for BCPNP Application model",
            "web_function": "✅ ",
        },
        "bcpnp-app-ee": {
            "path": "webform.bcpnp.bcpnpmodel_app",
            "class_list": ["BcpnpEEModelApp","BcpnpEEModelAppE"],
            "remark": "BCPNP module for BCPNP application (EE) model",
            "web_function": "✅ ",
        },
        # PR
        "0008": {
            "path": "model.pr.m0008",
            "class_list": ["M0008Model","M0008ModelE"],
            "remark": "PR module for form 0008 model",
        },
        "5406": {
            "path": "model.pr.m5406",
            "class_list": ["M5406Model","M5406ModelE"],
            "remark": "PR module for form 5406 model",
        },
        "5562": {
            "path": "model.pr.m5562",
            "class_list": ["M5562Model","M5562ModelE"],
            "remark": "PR module for form 5562 model",
        },
        "5669": {
            "path": "model.pr.m5669",
            "class_list": ["M5669Model","M5669ModelE"],
            "remark": "PR module for form 5669 model",
        },
        "pr": {
            "path": "webform.prportal.prmodel",
            "class_list": ["PrModel","PrModelE"],
            "pdf_function": "use prd",
            "remark": "PR module for all PR model",
        },
        # TR
        "1295": {
            "path": "model.tr.m1295",
            "class_list": ["M1295Model","M1295ModelE"],
            "pdf_function": "✅",
            "docx_template": {
                "sl": DATADIR / "word" / f"1295-sl-{rcic_company_id_name}.docx"
            },
            "remark": "TR module for form 1295 model",
        },
        "5257": {
            "path": "model.tr.m5257",
            "class_list": ["M5257Model","M5257ModelE"],
            "pdf_function": "✅",
            "remark": "TR module for form 5257 model",
        },
        "5708": {
            "path": "model.tr.m5708",
            "class_list": ["M5708Model","M5708ModelE"],
            "pdf_function": "✅",
            "docx_template": {
                "sl": DATADIR / "word" / f"5708-sl-{rcic_company_id_name}.docx"
            },
            "remark": "TR module for form 5708 model",
        },
        "5709": {
            "path": "model.tr.m5709",
            "class_list": ["M5709Model","M5709ModelE"],
            "pdf_function": "✅",
            "remark": "TR module for form 5709 model",
        },
        "5710": {
            "path": "model.tr.m5710",
            "class_list": ["M5710Model","M5710ModelE"],
            "pdf_function": "✅",
            "remark": "TR module for form 5710 model",
        },
        "5476": {
            "path": "model.common.m5476",
            "class_list": ["M5476Model","M5476ModelE"],
            "pdf_function": "✅ ",
            "remark": "Rep form",
            "help": {
                "description": "This model can automatically make json file for filling pdf form ",
                "helps": helps["5476"],
            },
        },
    }

    return models



from webform.webcommon import WebPages, Page


class LmBenefits(WebPages):
    @property
    def actions(self):
        return [self.page1, self.page2, self.page3]

    @property
    def page1(self):
        job_creation = (
            [
                self.web_element.areatextElement(
                    "#\\39 507 > textarea",
                    self.app.lmi.job_creation_benefit,
                    label="Details for job creation",
                )
            ]
            if self.app.lmi.job_creation_benefit
            else []
        )
        skill_transfer_id = (
            "#\\39 508 > div > input:nth-child(1)"
            if self.app.lmi.skill_transfer_benefit
            else "#\\39 508 > div > input:nth-child(4)"
        )
        skill_transfer = self.web_element.radioElement(
            skill_transfer_id, label="Skill transfer"
        )

        page_actions = [*job_creation, skill_transfer]
        next_page_tag = (
            "#\\39 510 > div > input:nth-child(4)"  # labour shortage ? No-id
        )

        return Page(page_actions, "#next", next_page_tag, label="Job creation").page

    @property
    def page2(self):
        skill_transfer = (
            [
                self.web_element.areatextElement(
                    "#\\39 509 > textarea",
                    self.app.lmi.skill_transfer_benefit,
                    label="Details for skill transfer",
                )
            ]
            if self.app.lmi.skill_transfer_benefit
            else []
        )
        labour_shortage_id = (
            "#\\39 510 > div > input:nth-child(1)"
            if self.app.lmi.skill_transfer_benefit
            else "#\39 510 > div > input:nth-child(4)"
        )
        labour_shortage = self.web_element.radioElement(
            labour_shortage_id, label="Skill transfer"
        )

        page_actions = [*skill_transfer, labour_shortage]
        next_page_tag = "#\\39 513 > div > input:nth-child(4)"  # lay off ? No-id

        return Page(page_actions, "#next", next_page_tag, label="Skill transfer").page

    @property
    def page3(self):
        labour_shortage = (
            [
                self.web_element.areatextElement(
                    "#\\39 511 > textarea",
                    self.app.lmi.fill_shortage_benefit,
                    label="Details for filling labour",
                )
            ]
            if self.app.lmi.fill_shortage_benefit
            else []
        )

        other_benefit = self.web_element.areatextElement(
            "#\\39 512 > textarea", self.app.lmi.other_benefit, label="Other benefits"
        )

        layoff_id = (
            "#\\39 513 > div > input:nth-child(1)"
            if self.app.lmi.laid_off_in_12
            else "#\\39 513 > div > input:nth-child(4)"
        )
        layoff = self.web_element.radioElement(
            layoff_id, label="Has laid off in past 12 months?"
        )

        page_actions = [*labour_shortage, other_benefit, layoff]
        next_page_tag = "#\\39 518 > div > input:nth-child(4)"  # hiring will lead to lay-offs ? No-id

        return Page(page_actions, "#next", next_page_tag, label="Skill transfer").page



from datetime import date
from pydantic import BaseModel, EmailStr
from typing import List, Optional
from datetime import date
from model.common.commonmodel import CommonModel, BuilderModel
from model.common.rcic import RcicList
from pdfform.common.fb5476 import FormBuilder5476
import json
from termcolor import colored


class Personal(BaseModel):
    last_name: str
    first_name: str
    sex: str
    dob: date
    uci: Optional[str]


class M5476Model(BaseModel, BuilderModel):
    rciclist: List[RcicList]
    personal: Personal

    def make_pdf_form(self, output_json, rcic_id_name, *args, **kwargs):
        pf = FormBuilder5476(self, rcic_id_name)
        form = pf.get_form()
        with open(output_json, "w") as output:
            json.dump(form.actions, output, indent=3, default=str)
        print(colored(f"{output_json} has been created. ", "green"))

    def make_web_form(self, output_json, upload_dir, rcic, *args, **kwargs):
        raise ValueError("This model doesn't have webform...")

    def context(self, *args, **kwargs):
        raise ValueError("This model doesn't have webform...")


class M5476ModelE(CommonModel, M5476Model):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/pa.xlsx",
            "excel/rep.xlsx",
        ]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from termcolor import colored
from functools import reduce
from typing import List, Optional
from model.common.address import Address
from model.common.phone import Phone
from model.common.trperson import (
    COR,
    PersonId,
    Personal,
    Marriage,
    Education,
    Employment,
    Travel,
    Family,
)
from model.common.tr import TrCase, Visa, TrBackground
from model.common.commonmodel import CommonModel
from pydantic import BaseModel

"""
Program model for temporary resident visa. Get and validate info for forms: imm5257, imm0104, imm5257b_1, and imm5645
"""


class M5257Model(BaseModel):
    personal: Personal
    marriage: Marriage
    personid: List[PersonId]
    address: List[Address]
    education: List[Education]
    employment: List[Employment]
    travel: List[Travel]
    family: List[Family]
    phone: List[Phone]
    cor: List[COR]
    trcase: TrCase
    visa: Visa
    trbackground: TrBackground


class M5257ModelE(CommonModel, M5257Model):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/tr.xlsx", "excel/pa.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



# from context import BASEDIR
import os, json, time, argparse, shutil
from pywinauto.application import Application
from pywinauto.keyboard import send_keys
from pdfform import config_win
from pdfform.application_form import ApplicationForm
from pdfform.form_controls_win import Skip
from pathlib import Path
from utils.utils import append_ext

# Get project's home directory,
BASEDIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
# All data directory
DATADIR = os.path.abspath(os.path.join(BASEDIR, "data"))


def check_env():
    if not os.path.isfile(config_win.ACROBAT_READER_PATH):
        print("Acrobat reader not found. Please install it first.")
        exit()


def prepare_document(form_type: str, name: str, wait_time: int):
    """Create a new empty pdf file to fill"""

    # copy empty template file to output folder
    src = Path(DATADIR + f"/pdf/{form_type}.pdf")
    dst = Path(".").cwd() / f"{name}"
    shutil.copy2(src, dst)

    # open file with acrobat reader
    app = Application(backend="uia").start(
        str(config_win.ACROBAT_READER_PATH) + " " + str(dst)
    )
    # TODO: find better way to detect ready
    time.sleep(
        wait_time and int(wait_time) or 10
    )  # wait 10 seconds for file open finished.


def finish_document():
    """Validate, save and close document"""

    send_keys("^S")  # Ctrl + S to save
    send_keys("^W")  # Ctrl + W to close


def get_args():
    """Get command-line arguments"""

    parser = argparse.ArgumentParser(
        description="PDF Filler -- automatically fill application forms.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("form_name", help="Input form name, for example: 1294")
    parser.add_argument("-j", "--json_name", help="Input json file name")
    parser.add_argument("-w", "--wait", help="wait time(seconds) for open pdf")
    parser.add_argument(
        "-o",
        "--output",
        help="Output file name. ",
        metavar="output_file_name",
        default="test",
    )

    parser.add_argument(
        "-v", "--verbose", help="verbose output filling steps", action="store_true"
    )
    args = parser.parse_args()
    args.output = append_ext(args.json_name, ".pdf")
    args.json_name = append_ext(args.json_name, ".json")

    return args


def main():
    """main function"""

    args = get_args()
    form_name = args.form_name
    in_file = str(Path(args.json_name))
    verbose = args.verbose
    out_file = str(Path(".").cwd() / args.output)

    check_env()
    prepare_document(form_name, out_file, args.wait)

    # open actions data
    with open(in_file) as f:
        actions = json.load(f)

    form = ApplicationForm(verbose=verbose)

    for action in actions:
        form.add_step(action)

    form.fill_form(0, verbose=verbose)
    finish_document()


if __name__ == "__main__":
    main()



from datetime import date
from typing import Union
from urllib.parse import ParseResultBytes
from pdfform.tr import option_lists
from pdfform.jsonmaker import JsonMaker
from model.common.cor import CORs
from model.common.id import ID, IDs
from model.common.address import Address, Addresses
from model.common.phone import Phone, Phones
from model.common.educationbase import EducationBase, EducationHistory


class TrCommon:
    def __init__(
        self,
        applicant: object,
    ):
        self.applicant = applicant
        self.form = JsonMaker()

    def special_YMD_handle(self, the_date):
        the_date = the_date.strftime("%Y-%m-%d") if type(the_date) == date else the_date
        year, month, day = the_date.split("-")
        self.form.add_text(year)
        self.form.add_text(month)
        self.form.add_text(day)

    def add_personal_detail(self, form_type: str = ""):
        """Add personal detail items
        1. full name
        2. other name
        3. sex
        4. date of birth
        5. place of birth
        6. citizenship
        7. current country or territory of residence
        8. previous countries or territories of residence
        9. country or territory where applying
        10. current marital status
        11. previous mariage
        """
        personal = self.applicant.personal
        self.form.add_info("personal detail section")

        self.form.add_text(personal.last_name)
        self.form.add_text(personal.first_name)

        if personal.used_first_name:
            self.form.add_radio(True)
            self.form.add_text(personal.used_last_name)
            self.form.add_text(personal.used_first_name)
        else:
            self.form.add_radio(False)
            self.form.add_skip(2)

        self.form.add_dropdown(personal.sex, option_lists.gender, like=True, pause=0.01)
        # self.form.add_skip(1)  # TODO: remove it after testing
        self.form.add_date(personal.dob, pause=0.2)

        self.form.add_text(personal.place_of_birth)
        self.form.add_dropdown(personal.country_of_birth, option_lists.country_of_birth)

        self.form.add_dropdown(
            personal.citizen, option_lists.country_of_citizen, like=True, pause=0.01
        )

        # residence
        residence = CORs(self.applicant.cor).current
        self.form.add_dropdown(residence.country, option_lists.country_of_residence)
        status = residence.status
        self.form.add_dropdown(status, option_lists.immigration_status)
        if status.lower() in ["visitor", "worker", "student"]:
            self.form.add_skip(1)
            self.form.add_date(residence.start_date)
            self.form.add_date(residence.end_date)
        elif status.lower() in ["other"]:
            self.form.add_text("")
            self.form.add_date(residence.start_date)
            self.form.add_date(residence.end_date)
        else:
            self.form.add_skip(3)

        # previous residence
        previous_cors = CORs(self.applicant.cor).previous
        if previous_cors and len(previous_cors) > 0:
            p1 = previous_cors[0]
            p2 = previous_cors[1] if len(previous_cors) > 1 else None

            self.form.add_radio(True)
            self.form.add_dropdown(p1.country, option_lists.country_of_residence)
            self.form.add_dropdown(p1.status, option_lists.immigration_status)
            self.form.add_skip(1)
            self.form.add_date(p1.start_date)
            self.form.add_date(p1.end_date)

            if p2:
                self.form.add_dropdown(p2.country, option_lists.country_of_residence)
                self.form.add_dropdown(p2.status, option_lists.immigration_status)
                self.form.add_skip(1)
                self.form.add_date(p2.start_date)
                self.form.add_date(p2.end_date)
            else:
                self.form.add_skip(5)
        else:
            self.form.add_radio(False)
            self.form.add_skip(10)

        if form_type == "1295":
            if not self.applicant.trcase.same_as_cor:
                self.form.add_radio(False)
                tr = self.applicant.trcase
                self.form.add_dropdown(
                    tr.applying_country, option_lists.previous_country
                )
                self.form.add_dropdown(
                    tr.applying_status, option_lists.immigration_status
                )
                if tr.applying_status == "Other":
                    self.form.add_text(tr.other_explain)
                else:
                    self.form.add_skip(1)
                self.form.add_date(tr.applying_start_date)
                self.form.add_date(tr.applying_end_date)

            else:
                self.form.add_radio(True)
                self.form.add_skip(5)

        marriage = self.applicant.marriage
        status = marriage.marital_status

        self.form.add_dropdown(status, option_lists.marital_status, True)
        if status.lower() in ["married", "common-law"]:
            self.form.add_date(marriage.married_date)
            self.form.add_text(marriage.sp_last_name)
            self.form.add_text(marriage.sp_first_name)
            if form_type != "1295":
                self.form.add_radio(marriage.sp_is_canadian)
        else:
            self.form.add_skip(3 if form_type == "1295" else 4)

        if marriage.previous_married:
            self.form.add_radio(True)
            self.form.add_text(marriage.pre_sp_last_name)
            self.form.add_text(marriage.pre_sp_first_name)
            if not marriage.pre_sp_dob:
                raise ValueError("Previous spouse dob is None")
            dob_list = marriage.pre_sp_dob.strftime("%Y-%m-%d").split("-")
            if form_type == "1295":
                self.form.add_text(dob_list, pause=0.1)
            self.form.add_dropdown(
                marriage.pre_relationship_type, option_lists.relation_type
            )
            self.form.add_date(marriage.pre_start_date)
            self.form.add_date(marriage.pre_end_date)
            # speciaially dealing with
            if form_type != "1295":
                self.special_YMD_handle(marriage.pre_sp_dob)

        else:
            self.form.add_radio(False)
            self.form.add_skip(8)

    def add_language(self):
        """Add language items`
        1. native language
        """
        self.form.add_info("language section")

        self.form.add_dropdown(
            self.applicant.personal.native_language, option_lists.native_language
        )
        self.form.add_dropdown(
            self.applicant.personal.english_french, option_lists.communication_language
        )
        if self.applicant.personal.english_french.lower() == "both":
            self.form.add_text(self.applicant.personal.which_one_better)
        else:
            self.form.add_skip(1)
        if self.applicant.personal.language_test:
            self.form.add_radio(True)
        else:
            self.form.add_radio(False)

    def add_passport(self):
        """Add passport items
        1. passport number
        2. issue country`
        3. issue date
        4. expiry date
        5. Taiwan
        6. Israeli
        """
        self.form.add_info("passport section")

        passport: ID = IDs(self.applicant.personid).passport
        self.form.add_text(passport.number)
        self.form.add_dropdown(
            passport.country, option_lists.passport_issue_country, True
        )
        self.form.add_date(passport.issue_date)
        self.form.add_date(passport.expiry_date)

        if "taiwan" in passport.country.lower():
            self.form.add_radio(True)
        else:
            self.form.add_skip(1)
        if "israel" in passport.country.lower():
            self.form.add_radio(True)
        else:
            self.form.add_skip(1)

    def add_national_id(self):
        """Add national id items
        1. if has national id
        2. id number
        3. issue country
        4. issue date
        5. expiry date
        """

        self.form.add_info("national id section")
        national_id: ID = IDs(self.applicant.personid).national_id
        if national_id.number:
            self.form.add_radio(True)
            self.form.add_text(national_id.number)
            self.form.add_dropdown(
                national_id.country, option_lists.passport_issue_country, True
            )
            if national_id.issue_date:
                self.form.add_date(national_id.issue_date)
            else:
                self.form.add_skip(1)
            if national_id.expiry_date:
                self.form.add_date(national_id.expiry_date)
            else:
                self.form.add_skip(1)
        else:
            self.form.add_radio(False)
            self.form.add_skip(4)

    def add_uspr_card(self):
        self.form.add_info("US PR card section")

        """Add US PR card items"""
        pr: ID = IDs(self.applicant.personid).pr
        if pr.country and (
            "united states" in pr.country.lower() or "usa" in pr.country.lower()
        ):
            self.form.add_radio(True)
            self.form.add_text(pr.number)
            self.form.add_date(pr.expiry_date)
        else:
            self.form.add_radio(False)
            self.form.add_skip(2)

    def add_contact_information(self, form_type: str = ""):
        """Add contact information"""

        self.form.add_info("contact information section")

        address: Address = Addresses(self.applicant.address).mailing
        residential_address: Address = Addresses(self.applicant.address).residential

        self.form.add_text(address.po_box)
        self.form.add_text(address.unit)
        self.form.add_text(address.street_number)
        self.form.add_text(address.street_name)
        self.form.add_text(address.city)
        country = address.country
        self.form.add_dropdown(country, option_lists.mailing_country)
        if country.lower() == "canada":
            self.form.add_dropdown(address.province, option_lists.canada_province)
        else:
            self.form.add_skip(1)
        self.form.add_text(address.post_code)
        if form_type == "1295":
            self.form.add_text(address.district)

        if residential_address != address:
            self.form.add_radio(False)
            self.form.add_text(residential_address.unit)
            self.form.add_text(residential_address.street_number)
            self.form.add_text(residential_address.street_name)
            self.form.add_text(residential_address.city)
            country = residential_address.country
            self.form.add_dropdown(country, option_lists.mailing_country)
            if country.lower() == "canada":
                self.form.add_dropdown(
                    residential_address.province, option_lists.canada_province
                )
            else:
                self.form.add_skip(1)
            self.form.add_text(residential_address.post_code)
            if form_type == "1295":
                self.form.add_text(residential_address.district)

        else:
            self.form.add_radio(True)
            if form_type == "1295":
                self.form.add_skip(8)
            else:
                self.form.add_skip(7)

        # telephone
        telephone: Union[Phone, None] = Phones(self.applicant.phone).PreferredPhone
        phone_type = (
            "Residence"
            if telephone.variable_type == "residential"
            else telephone.variable_type.title()
        )
        if telephone and telephone.isCanadaUs:
            pause = 3  # this place could be easily erred
            self.form.add_checkbox(True)
            self.form.add_skip(1)
            self.form.add_dropdown(phone_type, option_lists.telephone_type)
            self.form.add_skip(1)
            if form_type == "1295":  # try to work around 1295 tab order issue
                self.form.add_text(telephone.ext)
                # self.form.add_text(telephone.number, pause=pause)
                self.form.add_text(telephone.NA_format_list, pause=pause)
                self.form.add_skip(3)
            else:
                # when a valid Canada phone number finished, it will automatically tab to next control
                ext = telephone.ext if telephone.ext else ""
                self.form.add_text(telephone.number + ext, pause=1)

        else:
            pause = 0.1
            self.form.add_skip(1)
            self.form.add_checkbox(True)
            self.form.add_dropdown(phone_type, option_lists.telephone_type)
            self.form.add_text(telephone.country_code, pause=pause)
            self.form.add_text(telephone.number, pause=pause)
            self.form.add_text(telephone.ext, pause=pause)
        # alternative phone or fax is not used usually, ignore alternate telephone and fax
        self.form.add_skip(11)

        self.form.add_text(self.applicant.personal.email)

    def add_coming_into_canada(self):
        """Add coming into canada."""

        self.form.add_info("coming into Canada section")

        coming = self.applicant.trcasein
        self.form.add_date(coming.original_entry_date)
        self.form.add_text(coming.original_entry_place)
        self.form.add_dropdown(coming.original_purpose, option_lists.purpose_of_coming)

        if coming.original_purpose == "Other":
            self.form.add_text(coming.original_other_reason)
        else:
            self.form.add_skip(1)

        if coming.most_recent_entry_date:
            self.form.add_date(coming.most_recent_entry_date)
            self.form.add_text(coming.most_recent_entry_place)
            self.form.add_text(coming.doc_number)
        else:
            self.form.add_skip(3)

    def add_education(self):
        """Add education"""
        self.form.add_info("education section")

        education: Union[EducationBase, None] = EducationHistory(
            self.applicant.education
        ).highestEducation
        pause = 0.6  # tested for many times, the speed always works
        if education and education.level_in_num >= 2:  # >=2 means post secondary
            self.form.add_radio(True)
            self.form.add_date(education.start_date, True, pause=pause)
            self.form.add_text(education.field_of_study)
            self.form.add_text(education.school_name)
            self.form.add_date(education.end_date, True, pause=pause)
            self.form.add_text(education.city)
            country = education.country
            self.form.add_dropdown(country, option_lists.country_of_birth)
            if country and country.lower() == "canada":
                self.form.add_dropdown(education.province, option_lists.canada_province)
            else:
                self.form.add_skip(1)
        else:
            self.form.add_radio(False)
            self.form.add_skip(9)

    def add_employment(self, form_type: str = ""):
        """Add employment"""
        self.form.add_info("employment section")

        slots = 3  # num of slots in form
        pause = 0.5
        if self.applicant.employment:
            employments = self.applicant.employment
            for employment in employments:
                # WTF!! work around the form bug
                if slots == 2 and form_type == "1295":
                    year, month, day = employment.start_date.strftime("%Y-%m-%d").split(
                        "-"
                    )
                    self.form.add_text(year)
                    self.form.add_text(month)
                else:
                    self.form.add_date(employment.start_date, True, pause=pause)
                self.form.add_text(employment.job_title)
                self.form.add_text(employment.company)
                self.form.add_date(employment.end_date, True, pause=pause)
                self.form.add_text(employment.city)
                country = employment.country
                self.form.add_dropdown(country, option_lists.country_of_birth)
                if country.lower() == "canada":
                    self.form.add_dropdown(
                        employment.province, option_lists.canada_province
                    )
                else:
                    self.form.add_skip(1)
                slots -= 1
                if slots == 0:
                    break

        # skip emply slots
        for _ in range(slots):
            self.form.add_skip(9)

    def add_background(self):
        """Add background information"""
        self.form.add_info("background information section")
        self.form.add_skip(1)  # skip for clear section
        background: TrBackground = self.applicant.trbackground
        self.form.add_radio(background.q1a)
        self.form.add_radio(background.q1b)
        if background.q1a or background.q1b:
            self.form.add_text(background.q1c)
        else:
            self.form.add_skip(1)

        self.form.add_radio(background.q2a)
        self.form.add_radio(background.q2b)
        self.form.add_radio(background.q2c)
        if background.q2a or background.q2b or background.q2c:
            self.form.add_text(background.q2d)
        else:
            self.form.add_skip(1)

        self.form.add_radio(background.q3a)
        if background.q3a:
            self.form.add_text(background.q3b)
        else:
            self.form.add_skip(1)

        self.form.add_radio(background.q4a)
        if background.q4a:
            self.form.add_text(background.q4b)
        else:
            self.form.add_skip(1)

        self.form.add_radio(background.q5)

        self.form.add_radio(background.q6)

    def add_signature(self):
        """Add signature section"""
        self.form.add_radio(True)

        self.form.add_text(
            f"{self.applicant.personal.last_name} {self.applicant.personal.first_name}"
        )  # signature
        self.form.add_text(str(date.today()))



from .context import DATADIR
from email.policy import default
from textwrap import indent
from typing import List, Optional
from model.experience.resumedata import (
    Personal,
    Language,
    PersonalAssess,
    Education,
    Employment,
)
from model.common.commonmodel import CommonModel
from model.common.phone import Phone, Phones
from model.common.address import Address, Addresses
from pydantic import BaseModel


class ResumeModel(BaseModel):
    personal: Personal
    phone: List[Phone]
    personalassess: PersonalAssess
    education: List[Education]
    language: Optional[List[Language]]
    employment: Optional[List[Employment]]
    address: List[Address]

    def context(self, doc_type=None):
        return {
            **self.dict(),
            "phone": Phones(self.phone).PreferredPhone.international_format_full,
            "address": Addresses(self.address).PreferredAddress.full_address,
            "full_name": self.personal.full_name,
        }


class ResumeModelE(CommonModel, ResumeModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/pa.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



# import context
from datetime import date
from apps.api.login import get_session
import json
from apps.api.routers.case import permission_check
from source.excel import Excel
from pathlib import Path
import typer, json
from utils.utils import append_ext, remove_ext
from apps.models import get_models
from rich.console import Console
from rich.text import Text
from rich.table import Table
from termcolor import colored
from model.common.wordmaker import WordMaker
from apps.client.config import Config
from utils.utils import makeTable

# Get project's home directory,
BASEDIR = Path(__file__).parents[1]
# All data directory
DATADIR = BASEDIR / "data"

app = typer.Typer()
console = Console()


session = get_session()

headers = {"accept": "application/json", "Content-Type": "application/json"}


def getJsonData(excel_file_name):
    e = Excel(excel_name=excel_file_name)
    return json.loads(e.json)


def show_exception(e: Exception):
    console.print(e, style="red")


def show_error(e):
    console.print(e, style="red")


def show_warning(msg: str):
    console.print(msg, style="yellow")


def getModel(model_name: str):
    # get model, and its instance
    models = get_models(rcic_company_id_name="", temp_num=1)
    model = models.get(model_name)
    if not model:
        show_error(f"{model_name} is not a valid model name.")
        exit(1)
    class_list = model["class_list"]
    the_class = __import__(model["path"], fromlist=class_list)
    the_model = getattr(the_class, class_list[1])  # No 2 is the Excel model
    return the_model


@app.command()
def man(model: str = typer.Argument(None, help="A model name, such as 5708")):
    """Manual of imm app"""
    models = get_models(rcic_company_id_name="", temp_num=1)
    # give specific manu for a model
    if model:
        if model not in models:
            show_error(
                f"{model} is not valid in defined models. Please use 'imm man' to check all models"
            )
            return
        help = models.get(model).get("help")
        if help:
            # print(f"{help.get('description')}")
            # print(tabulate(help.get("helps"), tablefmt="fancy_grid"))
            table = makeTable(f"Help for model {model}", help.get("helps"))
            console.print(table)

        else:
            print("There is no help info provided yet.")
    else:

        table = Table(title="Models List")
        table.add_column("Model", justify="left", style="cyan", no_wrap=True)
        table.add_column("Word", style="magenta")
        table.add_column("Pdf Form", style="magenta")
        table.add_column("Web Form ", style="magenta")
        table.add_column("Description", justify="left", style="green")

        contents = []
        for model, value in models.items():
            word_template = value.get("docx_template")
            pdf_function = value.get("pdf_function")
            web_function = value.get("web_function")
            if word_template:
                temp_name = ", ".join(word_template)
            else:
                temp_name = ""
            contents.append(
                [
                    model,
                    Path(temp_name).name,
                    pdf_function,
                    web_function,
                    value["remark"],
                ]
            )
        print(
            "Every model can do two things. 1. -m make excel document 2. -c check excel content based on the model\nFor specific help, enter mmc model_name to get the model's details.",
        )
        for content in contents:
            table.add_row(*content)

        console.print(table)


@app.command()
def make(model: str, output_excel_name: str):
    """Make excel based on model"""
    output_excel_name = append_ext(output_excel_name, ".xlsx")
    file_exists = Path(output_excel_name).exists()
    if file_exists:
        overwrite = typer.confirm(f"{output_excel_name} is existed, overwrite?")
        if not overwrite:
            return
    # create excel
    request = {
        "model": model,
    }
    r = session.post("http://localhost:8000/case/make", json=request, headers=headers)
    if r.status_code == 401:
        print(r.json().get("detail"))
        return

    with open(output_excel_name, "wb") as f:
        f.write(r.content)
    print(colored(f"{output_excel_name} has been created", "green"))


@app.command()
def check(model: str, excel_name: str, print_json: bool = False):
    """Check input excel based on model"""
    excel_name = append_ext(excel_name, ".xlsx")
    request = {
        "model": model,
        "data": getJsonData(excel_name),
    }
    r = session.post("http://localhost:8000/case/check", json=request, headers=headers)
    if r.status_code == 401:
        print(r.json().get("detail"))
        return
    if r.status_code == 200:
        print(r.json()["success"])
    elif r.status_code == 422:
        print(r.json()["detail"])


@app.command()
def run(model: str, excel_name: str):
    """Run app only"""
    excel_name = append_ext(excel_name, ".xlsx")
    request = {
        "model": model,
        "data": getJsonData(excel_name),
    }
    r = session.post("http://localhost:8000/case/run", json=request)

    if r.status_code == 200:
        print(r.json()["success"])
    elif r.status_code == 422:
        print(r.json()["detail"])
    elif r.status_code == 401:
        print(r.json().get("detail"))
        return


@app.command()
def word(
    model: str,
    excel_name: str,
    word_name: str = typer.Argument(
        None,
        help="Output word name. If none, willl use excel's name without ext, plus .docx",
    ),
    doctype: str = typer.Argument(
        None,
        help="Which document to generate, such as sl, ert,eet,... Use 'imm man' to check word column for details",
    ),
    rciccompany: str = typer.Option(
        "noah", help="Rcic company short name defined in docx templates"
    ),
    tempnum: int = typer.Option(1, help="template number"),
):
    """Make word based on model and input excel"""
    # check
    excel_name = append_ext(excel_name, ".xlsx")
    if word_name:
        word_name = append_ext(word_name, ".docx")
    else:
        word_name = remove_ext(excel_name) + ".docx"

    word_file_exists = Path(word_name).exists()
    excel_file_exists = Path(excel_name).exists()

    if not excel_file_exists:
        print(colored(f"{excel_name} is not existed."))
        return

    if word_file_exists:
        overwrite = typer.confirm(f"{word_name} is existed, overwrite?")
        if not overwrite:
            return

    models = get_models(rcic_company_id_name=rciccompany, temp_num=tempnum)
    template_docx_dict = models.get(model).get("docx_template")
    if not template_docx_dict:
        print(
            colored(
                f"There is no template existing in model {model.get('class_list')[0]}",
                "red",
            )
        )
        return

    request = {
        "model": model,
        "data": getJsonData(excel_name),
        "output_name": word_name,
        "doctype": doctype,
        "rciccompany": rciccompany,
        "tempnum": tempnum,
    }
    r = session.post("http://localhost:8000/case/word", json=request)
    with open(word_name, "wb") as f:
        f.write(r.content)
    print(f"{word_name} has been downloaded from web")


@app.command()
def webform(
    model: str,
    excel_name: str,
    outputjson: str = typer.Argument(
        None,
        help="Output json file name. Without input, excel's name will be used  by adding '.json'",
    ),
    rcic: str = typer.Option(Config.rcic, help="RCIC's short name"),
    initial: bool = typer.Option(
        True, help="Is this for starting a new case. Only used for BCPNP webform."
    ),
    previous: bool = typer.Option(
        False, help="Does this client have previous BCPNP application?"
    ),
    uploaddir: str = typer.Option(
        ".", help="A directory in which all files will be uploaded."
    ),
):
    """Make json file based on model and input excel for webform filling"""
    excel_name = append_ext(excel_name, ".xlsx")
    if outputjson:
        outputjson = append_ext(outputjson, ".json")
    else:
        outputjson = remove_ext(excel_name) + ".json"
    if not uploaddir:
        show_warning(
            "You did not input uploading directory. Without it , the files in current folder will be uploaded."
        )

    request = {
        "model": model,
        "data": getJsonData(excel_name),
        "output_name": outputjson,
        "rcic": rcic,
        "upload_dir": uploaddir,
    }
    r = session.post("http://localhost:8000/case/webform", json=request)
    with open(outputjson, "wb") as f:
        f.write(r.content)
    print(f"{outputjson} has been downloaded from web")


@app.command()
def pdfform(
    model: str,
    excel_name: str,
    outputjson: str = typer.Argument(
        None,
        help="Output json file name. Without input, excel's name will be used  by adding '.json'",
    ),
    rcic: str = typer.Option(Config.rcic, help="RCIC's short name"),
):
    """Make json file based on model and input excel for pdf form filling"""

    excel_name = append_ext(excel_name, ".xlsx")

    if outputjson:
        outputjson = append_ext(outputjson, ".json")
    else:
        outputjson = remove_ext(excel_name) + ".json"

    request = {
        "model": model,
        "data": getJsonData(excel_name),
        "output_name": outputjson,
        "rcic": rcic,
    }
    r = session.post("http://localhost:8000/case/pdfform", json=request)
    with open(outputjson, "wb") as f:
        f.write(r.content)
    print(f"{outputjson} has been downloaded from web")


if __name__ == "__main__":
    app()



import context
from utils.language import CELPIP, IELTS, TEF, TCF, CLB
from datetime import date


def show(clb):
    print(
        clb.clb,
        clb.clb_r,
        clb.clb_w,
        clb.clb_s,
        clb.clb_l,
        clb.is_valid(),
        clb.test_date,
        clb.sign_date,
    )


ielts = {
    "reading": 6.5,
    "writting": 6.5,
    "speaking": 7,
    "listening": 6.5,
    "test_date": date.today(),
    "sign_date": date.today(),
    "type_of_test": "Academic",
    "report_number": "112233",
}
clb = IELTS(**ielts)

show(clb)
print(clb.type_of_test, clb.report_number)

celpip = {
    "reading": 6.5,
    "writting": 6,
    "speaking": 7,
    "listening": 6,
    "test_date": date.today(),
    "sign_date": date.today(),
    "registration_number": "123321",
    "pin_number": "112233",
}
clb = CELPIP(**celpip)
show(clb)
print(clb.registration_number, clb.pin_number)

tef = {
    "reading": 266,
    "writting": 333,
    "speaking": 288,
    "listening": 333,
    "test_date": date.today(),
    "sign_date": date.today(),
}
clb = TEF(**tef)
show(clb)

tcf = {
    "reading": 344,
    "writting": 20,
    "speaking": 20,
    "listening": 333,
    "test_date": date.today(),
    "sign_date": date.today(),
}
clb = TCF(**tcf)
show(clb)

clb = CLB(7)
print(clb.to_ielts(), clb.to_tcf())



from model.common.advertisement import Advertisements, InterviewRecords
from model.common.finance import Finance
from model.common.advertisement import Advertisement, Advertisements
from model.common.advertisement import (
    InterviewRecord,
    InterviewRecords,
    RecruitmentSummary,
)
from model.common.contact import Contacts
from model.lmia.common import *
from model.common.commonmodel import CommonModel, BuilderModel
from webform.lmiaportal.webformmodel import WebformModel
from pydantic import BaseModel


class M5593Model(BaseModel, BuilderModel):
    lmiacase: LmiaCase
    general: General
    lmi: Lmi
    eraddress: List[ErAddress]
    contact: List[Contact]
    finance: List[Finance]
    joboffer: Joboffer
    personal: Personal
    personalassess: PersonalAssess
    position: Position
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    rcic: Rcic
    recruitmentsummary: RecruitmentSummary

    def context(self, *args, **kwargs):
        interview = InterviewRecords(self.interviewrecord)
        primary_contact = Contacts(self.contact).primary
        return {
            **self.dict(),
            "formatted_finance": [
                {
                    "year": f.year,
                    "formatted_revenue": f.formatted_revenue,
                    "formatted_net_income": f.formatted_net_income,
                    "formatted_retained_earning": f.formatted_retained_earning,
                }
                for f in self.finance
            ],
            "num_of_job_posts": Advertisements(self.advertisement).amount,
            "why_qualified": self.personalassess.why_qualified,
            "adv_summary": [
                {
                    "start_date": a.start_date,
                    "end_date": a.end_date,
                    "media": a.media,
                    "days": a.days,
                }
                for a in self.advertisement
            ],
            "high_wage": True
            if float(self.joboffer.hourly_rate) >= self.lmiacase.provincial_median_wage
            else False,
            "rs": {
                "resume_num": interview.resume_num,
                "canadian_num": interview.canadian_num,
                "unknown_num": interview.unknown_num,
                "foreigner_num": interview.foreigner_num,
                "total_canadian": interview.total_canadian,
                "total_interviewed_canadians": interview.total_interviewed_canadians,
                "canadian_records": interview.canadian_records,
            },
            "primary_contact": primary_contact,
            "employee_name": self.personal.full_name,
        }

    def make_web_form(self, output_json, upload_dir, rcic, *args, **kwargs):
        args = {
            "model_variable": "5593",
            "app": self,
            "output_json": output_json,
            "upload_dir": upload_dir,
            "rcic": rcic,
        }
        wf = WebformModel(**args)
        wf.save()

    def make_pdf_form(self):
        pass


class M5593ModelE(CommonModel, M5593Model):
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/recruitment.xlsx",
            "excel/pa.xlsx",
            "excel/er.xlsx",
            "excel/rep.xlsx",
            "excel/lmia.xlsx",
        ]
        # call parent class for validating
        super().__init__(excels, output_excel_file, mother_excels, globals())



from pydantic import BaseModel
from typing import List
from model.common.commonmodel import CommonModel

# from context import DATADIR


class CapTFW(BaseModel):
    employee: str
    hourly_rate: float
    hours_per_week: float
    is_working: bool
    designated_position: bool
    pr_support_only_lmia: bool
    in_application: bool
    pr_in_process: bool


class General(BaseModel):
    canadian_ft_employee_num: int
    canadian_pt_employee_num: int


class LmiaCase(BaseModel):
    provincial_median_wage: float


class CapModel(BaseModel):
    general: General
    lmiacase: LmiaCase
    captfw: List[CapTFW]

    @property
    def A(self):
        ft_tfws_num = len(
            [tfw for tfw in self.captfw if tfw.hours_per_week >= 30 and tfw.is_working]
        )
        return self.general.canadian_ft_employee_num + ft_tfws_num

    @property
    def B(self):
        pt_tfws_num = len(
            [tfw for tfw in self.captfw if tfw.hours_per_week < 30 and tfw.is_working]
        )
        return self.general.canadian_pt_employee_num + pt_tfws_num

    @property
    def C(self):
        return len(
            [
                tfw
                for tfw in self.captfw
                if tfw.hourly_rate < self.lmiacase.provincial_median_wage
                and tfw.is_working
                and not tfw.in_application
            ]
        )

    @property
    def D(self):
        return len(
            [
                tfw
                for tfw in self.captfw
                if tfw.hourly_rate < self.lmiacase.provincial_median_wage
                and tfw.is_working
                and tfw.in_application
            ]
        )

    @property
    def E(self):
        return len(
            [
                tfw
                for tfw in self.captfw
                if tfw.hourly_rate < self.lmiacase.provincial_median_wage
                and not tfw.is_working
            ]
        )

    @property
    def F(self):
        return len(
            [
                tfw
                for tfw in self.captfw
                if tfw.hourly_rate < self.lmiacase.provincial_median_wage
                and tfw.pr_in_process
            ]
        )

    @property
    def G(self):
        if self.F == 0:
            return 0
        g = self.F / (self.C + self.D)
        return "{:,.2f}".format(g)

    @property
    def H(self):
        h = float(self.G) * (self.C + self.D - self.F)
        return "{:,.2f}".format(h)

    @property
    def step1(self):
        step = (self.C + self.D + self.E) - (self.F + float(self.H))
        return "{:,.2f}".format(step)

    @property
    def step2(self):
        if self.A + self.B / 2 < 10:
            return 10
        step = self.A + self.B / 2 + self.E
        return "{:,.2f}".format(step)

    @property
    def step3(self):
        if float(self.step1) == 0:
            return 0
        step = float(self.step1) / float(self.step2)
        return "{:,.2f}".format(step)

    @property
    def step4(self):
        step = float(self.step3) * 100
        return "{:,.2f}".format(step)

    @property
    def result(self):
        return {
            "A": self.A,
            "B": self.B,
            "C": self.C,
            "D": self.D,
            "E": self.E,
            "F": self.F,
            "G": self.G,
            "H": self.H,
            "step1": self.step1,
            "step2": self.step2,
            "step3": self.step3,
            "step4": self.step4,
        }


class CapModelE(CommonModel, CapModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/er.xlsx", "excel/lmia.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from webform.bcpnp.register import Register
from webform.bcpnp.bcpnpmodel_reg import BcpnpModelReg, BcpnpEEModelReg
from webform.bcpnp.bcpnpmodel_app import BcpnpEEModelApp, BcpnpModelApp
from webform.bcpnp.login import Login
from webform.bcpnp.registrant import Registrant
from webform.bcpnp.education_reg import EducationReg
from webform.bcpnp.employment import EmploymentReg
from webform.bcpnp.joboffer_reg import JobofferReg
from webform.bcpnp.language import LanguageReg
from webform.bcpnp.rep import Representative
from webform.bcpnp.submit import Submit
from pathlib import Path
from abc import ABC, abstractmethod
from webform.bcpnp.applicant import Applicant
from webform.bcpnp.education_app import EducationApp
from webform.bcpnp.workexperience import WorkExperience
from webform.bcpnp.family import FamilyApp
from webform.bcpnp.joboffer_app import JobofferApp
from webform.bcpnp.rep import Representative
from model.bcpnp.mrep_e import RcicList
from model.bcpnp.mrep_e import MRepModel


class Builder(ABC):
    def __init__(self, excels) -> None:
        if len(excels) > 0 and not Path(excels[0]).exists():
            raise FileExistsError(f"{excels[0]} is not existed")

    @abstractmethod
    def make_web_form(self, *args, **kwargs):
        raise NotImplementedError("This class must be implemented. ")


class ProfileBuilder(Builder):
    def __init__(self, excels, model: object):
        super().__init__(excels)
        self.applicant = model(excels=excels)

    def make_web_form(self):
        # signing in, pick skill immigration, select stream and confirm
        actions = Login(self.applicant).login()
        # Registrant
        actions += Registrant(self.applicant).fill()
        return actions


class RegBuilder(Builder):
    action_blocks = [
        Registrant,
        EducationReg,
        EmploymentReg,
        JobofferReg,
        LanguageReg,
        Submit,
    ]

    def __init__(self, excels, is_ee: bool = False):
        super().__init__(excels)
        self.applicant = (
            BcpnpEEModelReg(excels=excels) if is_ee else BcpnpModelReg(excels=excels)
        )

    def make_web_form(self, initial=True, previous=False):
        actions = Login(self.applicant).login(initial=initial, previous=previous)
        for block in RegBuilder.action_blocks:
            actions += block(self.applicant).fill()

        return actions


class AppBuilder(Builder):
    action_blocks = [
        Applicant,
        EducationApp,
        WorkExperience,
        FamilyApp,
        JobofferApp,
        Submit,
    ]

    def __init__(self, excels, is_ee: bool = False):
        super().__init__(excels)
        self.applicant = (
            BcpnpEEModelApp(excels=excels) if is_ee else BcpnpModelApp(excels=excels)
        )

    def make_web_form(self, initial=True, previous=False):
        actions = Login(self.applicant).login(initial=initial, previous=previous)
        for block in AppBuilder.action_blocks:
            actions += block(self.applicant).fill()

        return actions



from datetime import date
from typing import Optional, List
from pydantic import BaseModel, root_validator
from model.common.commonmodel import CommonModel
import os


class General(BaseModel):
    legal_name: str
    operating_name: Optional[str]
    website: Optional[str]
    establish_date: date
    num_pnps: int
    num_pnps_approved: int
    num_pnps_in_process: int
    has_lmia_approved: bool
    when_lmia_approved: Optional[date]
    last_revenue: float
    last_profit: float
    before_last_revenue: float
    before_last_profit: float
    retained_earning: float
    has_jobbank_account: bool
    has_bc_employer_certificate: bool


class Employee_list(BaseModel):
    employee: str
    job_title: str
    wage: float
    hours_per_week: float
    immigration_status: str
    employment_start_date: date
    remark: Optional[str]


class Lmi(BaseModel):
    laid_off_in_12: bool
    laid_off_canadians: Optional[int]
    laid_off_tfw: Optional[int]
    laid_off_reason: Optional[str]
    is_work_sharing: bool
    work_sharing_info: Optional[str]
    labour_dispute: bool
    labour_dispute_info: Optional[str]

    @root_validator
    def checkLayoff(cls, values):
        laid_off_in_12 = values.get("laid_off_in_12", None)
        laid_off_canadians = values.get("laid_off_canadians", None)
        laid_off_tfw = values.get("laid_off_tfw", None)
        laid_off_reason = values.get("laid_off_reason", None)
        if laid_off_in_12 and (
            not laid_off_canadians or not laid_off_tfw or not laid_off_reason
        ):
            raise ValueError(
                "Since there is laid of in past 12 months in info lmi sheet,but did not input how many Canadians and/or foreign workers, and/or reason of lay off."
            )
        return values

    @root_validator
    def checkWorkSharing(cls, values):
        is_work_sharing = values.get("is_work_sharing", None)
        work_sharing_info = values.get("work_sharing_info", None)
        if is_work_sharing and not work_sharing_info:
            raise ValueError(
                "Since there is work sharing in info lmi sheet,but did not input the details about it."
            )
        return values


class LmiaAssess(BaseModel):
    general: General
    employee_list: List[Employee_list]
    lmi: Lmi


class LmiaAssessE(CommonModel):
    general: General
    employee_list: List[Employee_list]
    lmi: Lmi

    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/er.xlsx", "excel/lmia.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from webform.webcommon import WebPages, Page
import glob


class Upload(WebPages):
    directory: str

    def __init__(self, app: object, directory):
        super().__init__(app)
        self.directory = directory

    @property
    def actions(self):
        return [self.page1]

    @property
    def page1(self):
        recruit_id = "#RecruitmentEffort"
        other_id = "#Other"
        file_list = glob.glob(self.directory + "/*")

        actions = []
        file_list.sort()
        for file in file_list:
            if self.app.lmiacase.stream_of_lmia != "LWS" "recruit" in file.lower():
                self.upload(recruit_id, actions, file)
                actions.append(self.click_upload)
            else:
                self.upload(other_id, actions, file)
                actions.append(self.click_upload)

        next_page_tag = None  # "#attestation"
        return Page(actions, "#summary", next_page_tag, label="upload documents").page

    def upload(self, id, actions, file):
        upload = self.web_element.uploadElement(id, file, label=file)
        actions.append((upload))

    @property
    def click_upload(self):
        return self.web_element.buttonElement("#uploadDocument", label="Upload button")



from webform.webcommon import WebPages, Page
from model.common.advertisement import Advertisements, InterviewRecords
from model.common.finance import Finances


class Recruitment(WebPages):
    """
    1. variation?
        yes->reason + provide detais of recruitment?    Yes -> Use job bank?    Yes/No->recruitment summary+lmi
        No->attempted recruit?  Yes->Use jobbank?   Yes/NO->recruitment summary+lmi
                                                    No ->reason+lmi
    """

    @property
    def actions(self):
        return [self.page1, *self.page2]

    @property
    def page1(self):
        has_variation_id = (
            "#\\39 484 > div > input:nth-child(1)"
            if self.app.lmiacase.is_waived_from_advertisement
            else "#\\39 484 > div > input:nth-child(4)"
        )
        page_actions = [
            self.web_element.radioElement(
                has_variation_id,
                label="Is the position subject to a variation in advertisement",
            )
        ]
        # based on yes or no variation
        next_page_tag = (
            "#\\39 486 > textarea"
            if self.app.lmiacase.is_waived_from_advertisement
            else "#\\39 488 > div > input:nth-child(4)"
        )
        return Page(
            page_actions,
            "#next",
            next_page_tag,
            label="Page 1 of Advertisement variation",
        ).page

    @property
    def page2(self):
        if self.app.lmiacase.is_waived_from_advertisement:
            sub_p1 = self.variation_yes()
            if self.app.lmiacase.provide_details_even_waived:
                sub_p2 = self.use_jobbank()
                sub_p3 = self.use_jobbank_follower()
                return [sub_p1, sub_p2, sub_p3]
            else:
                return [sub_p1, *self.lmi]
        else:
            sub_p1 = self.attempted_to_hire()
            # if not waived from advertisement, attemp to hire is must, so we do not check and go directly yes
            sub_p2 = self.use_jobbank()
            sub_p3 = self.use_jobbank_follower()
            return [sub_p1, sub_p2, sub_p3]

    def use_jobbank_follower(self):
        if self.app.lmiacase.use_jobbank:
            jobbank = Advertisements(self.app.advertisement).jobbank
            post_id = jobbank.advertisement_id if jobbank else ""
            isjobbank = self.web_element.inputElement(
                "#\\39 491 > input", post_id, label="Job post Id"
            )
            sub_p3_page_actions = [isjobbank, *self.recruiment_summary, *self.lmi]
        else:
            # explain why not using jobbank
            why_not_jobbank = self.web_element.areatextElement(
                "#\\39 492 > textarea",
                self.app.lmiacase.reason_not_use_jobbank,
                label="Why not using jobbank",
            )
            sub_p3_page_actions = [
                why_not_jobbank,
                *self.recruiment_summary,
                *self.lmi,
            ]
        next_page_tag = "#\\39 508 > div > input:nth-child(4)"  # Skill and knowledge transfer? No-> Id
        sub_p3 = self.web_element.pageElement(
            "#next",
            next_page_tag,
            actions=sub_p3_page_actions,
            label="Recruitment summary and LMI",
        )

        return sub_p3

    def attempted_to_hire(self):
        # if not variation, ask for attempted to recruit. It is must, so only yes here
        have_attemped = "#\\39 488 > div > input:nth-child(1)"
        page_actions = [self.web_element.radioElement(have_attemped, label="Yes")]
        next_page_tag = "#\\39 490 > div > input:nth-child(4)"  # using job bank No-id
        return Page(
            page_actions,
            "#next",
            next_page_tag,
            label="have you attempted to recruit Canadians",
        ).page

    def variation_yes(self):
        rational = self.web_element.areatextElement(
            "#\\39 486 > textarea",
            self.app.lmiacase.reason_for_waived,
            label="Creteria and reason for waiving advertisement",
        )
        provide_details_id = (
            "#\\39 487 > div > input:nth-child(1)"
            if len(self.app.interviewrecord) > 0
            else "#\\39 487 > div > input:nth-child(4)"
        )
        will_provide_details = self.web_element.radioElement(
            provide_details_id,
            label="In addition to applying for the variation, would provide recruitment details",
        )
        page_actions = [rational, will_provide_details]
        next_page_tag = "#\\39 490 > div > input:nth-child(4)"  # Using job bank? No id

        return Page(
            page_actions,
            "#next",
            next_page_tag,
            label="Page 2 of advertisement variation",
        ).page

    def use_jobbank(self):
        using_jobbank_id = (
            "#\\39 490 > div > input:nth-child(1)"
            if self.app.lmiacase.use_jobbank
            else "#\\39 490 > div > input:nth-child(4)"
        )
        next_page_tag2 = "#\\39 506 > div > input:nth-child(4)"  # Will hiring a TFW result in direct job creation -> NO id

        # use job bank?
        return Page(
            [self.web_element.radioElement(using_jobbank_id, label="Using job bank?")],
            "#next",
            next_page_tag2,
            label="Page 2.1 Using job bank",
        ).page

    @property
    def recruiment_summary(self):
        summary = InterviewRecords(self.app.interviewrecord)
        rs = [
            self.web_element.inputElement(
                "#\\39 494 > input",
                str(summary.canadian_num),
                label="number of Canadians",
            ),
            self.web_element.inputElement(
                "#\\39 495 > input",
                str(summary.total_interviewed_canadians),
                label="Total Canadians interviewed",
            ),
            self.web_element.inputElement(
                "#\\39 496 > input",
                str(summary.total_offered_canadians),
                label="Total Canadians offered",
            ),
            self.web_element.inputElement(
                "#\\39 497 > input",
                str(summary.total_hired_canadians),
                label="Total Canadians hired",
            ),
            self.web_element.inputElement(
                "#\\39 498 > input",
                str(summary.total_declined_joboffer_canadians),
                label="Total Canadians declined offer",
            ),
            self.web_element.inputElement(
                "#\\39 499 > input",
                str(summary.canadian_applied_not_interviewed_or_offered),
                label="Total Canadians applied but not interviewed or offered the job",
            ),
            self.web_element.areatextElement(
                "#\\39 500 > textarea",
                str(summary.why_canadians_not_hired),
                label="Why Canadian not hired",
                length=4000,
            ),
        ]

        return rs

    @property
    def lmi(self):
        """Labour Market Impacts"""
        employee_number = self.web_element.inputElement(
            "#\\39 502 > input", str(self.app.general.ft_employee_number)
        )
        last_revenue = Finances(self.app.finance).last_income or 0
        revenue_more_than_5m_id = (
            "#\\39 503 > div > input:nth-child(1)"
            if last_revenue >= 5000000
            else "#\\39 503 > div > input:nth-child(4)"
        )
        revenue_more_than_5m = self.web_element.radioElement(
            revenue_more_than_5m_id, label="Revenue more than 5M"
        )

        job_creation_id = (
            "#\\39 506 > div > input:nth-child(1)"
            if self.app.lmi.job_creation_benefit
            else "#\\39 506 > div > input:nth-child(4)"
        )
        job_creation = self.web_element.radioElement(
            job_creation_id, label="Job creation"
        )
        return [employee_number, revenue_more_than_5m, job_creation]



from webform.webcommon import WebPages, Page


class EmpBenefits(WebPages):
    @property
    def actions(self):
        jo = self.app.joboffer
        has_benefit = any(
            [
                jo.disability_insurance,
                jo.dental_insurance,
                jo.empolyer_provided_persion,
                jo.extended_medical_insurance,
                jo.extra_benefits,
            ]
        )
        return (
            [self.page1, self.page2, self.page3]
            if has_benefit
            else [self.page1, self.page21]
        )

    @property
    def page1(self):
        jo = self.app.joboffer
        has_benefit = any(
            [
                jo.disability_insurance,
                jo.dental_insurance,
                jo.empolyer_provided_persion,
                jo.extended_medical_insurance,
                jo.extra_benefits,
            ]
        )
        has_benefit_id = (
            "#\\39 471 > div > input:nth-child(1)"
            if has_benefit
            else "#\\39 471 > div > input:nth-child(4)"
        )
        page_actions = [
            self.web_element.radioElement(
                has_benefit_id, label="Will you provide benefits to the TFW?"
            )
        ]
        # Will you be providing any other -> No id, if has benefit
        # employement benefits pay percentage #\39 481 > input
        next_page_tag = (
            "#\\39 476 > div > input:nth-child(4)"
            if has_benefit
            else "#\\39 481 > input"
        )
        return Page(
            page_actions, "#next", next_page_tag, label="Page 1: Employment benefits"
        ).page

    @property
    def page2(self):
        jo = self.app.joboffer
        disability_insurance_id = (
            "#\\39 472 > div > input:nth-child(1)"
            if jo.disability_insurance
            else "#\\39 472 > div > input:nth-child(4)"
        )
        dental_insurance_id = (
            "#\\39 473 > div > input:nth-child(1)"
            if jo.dental_insurance
            else "#\\39 473 > div > input:nth-child(4)"
        )
        empolyer_provided_persion_id = (
            "#\\39 474 > div > input:nth-child(1)"
            if jo.empolyer_provided_persion
            else "#\\39 474 > div > input:nth-child(4)"
        )
        extended_medical_insurance_id = (
            "#\\39 475 > div > input:nth-child(1)"
            if jo.extended_medical_insurance
            else "#\\39 475 > div > input:nth-child(4)"
        )
        extra_benefits_id = (
            "#\\39 476 > div > input:nth-child(1)"
            if jo.extra_benefits
            else "#\\39 476 > div > input:nth-child(4)"
        )
        page_actions = [
            self.web_element.radioElement(disability_insurance_id, label="Disability"),
            self.web_element.radioElement(dental_insurance_id, label="Dental"),
            self.web_element.radioElement(
                empolyer_provided_persion_id, label="Pension"
            ),
            self.web_element.radioElement(
                extended_medical_insurance_id, label="Extened medical"
            ),
            self.web_element.radioElement(extra_benefits_id, label="Other"),
        ]
        next_page_tag = "#\\39 481 > input"  # Percentage of vacation pay id
        return Page(
            page_actions, "#next", next_page_tag, label="Page 2 of employment benefits"
        ).page

    @property
    def page3(self):
        page_actions = (
            [
                self.web_element.areatextElement(
                    "#\\39 477 > textarea",
                    self.app.joboffer.extra_benefits,
                    label="Extra benefits",
                )
            ]
            if self.app.joboffer.extra_benefits
            else []
        )

        number_paid_vacation = self.web_element.inputElement(
            "#\\39 480 > input",
            str(self.app.joboffer.vacation_pay_days),
            label="Paid vacation days",
        )
        percetage = self.web_element.inputElement(
            "#\\39 481 > input",
            str(self.app.joboffer.vacation_pay_percentage),
            label="Percentage of vacation pay",
        )

        page_actions.extend([number_paid_vacation, percetage])

        next_page_tag = "#\\39 484 > div > input:nth-child(4)"
        return Page(
            page_actions, "#next", next_page_tag, label="Page 3 of employment benefits"
        ).page

    @property
    def page21(self):
        page_actions = []

        number_paid_vacation = self.web_element.inputElement(
            "#\\39 480 > input",
            str(self.app.joboffer.vacation_pay_days),
            label="Paid vacation days",
        )
        percetage = self.web_element.inputElement(
            "#\\39 481 > input",
            str(self.app.joboffer.vacation_pay_percentage),
            label="Percentage of vacation pay",
        )

        page_actions.extend([number_paid_vacation, percetage])

        next_page_tag = "#\\39 484 > div > input:nth-child(4)"
        return Page(
            page_actions, "#next", next_page_tag, label="Page 3 of employment benefits"
        ).page



import time
from abc import ABC, abstractmethod
from typing import Tuple
from utils.utils import best_match
from pyautogui import press, write


class Control(ABC):
    """Pdf form input control base class"""

    verbose: bool = False

    @abstractmethod
    def fill(self):
        """Fill this control"""


class Skip(Control):
    """Skip disabled control"""

    def __init__(self, times=1, pause=0):
        self.pause = pause
        self.times = times

    def fill(self):
        if self.verbose:
            print(f"Skip to next for {self.times} times")
        for i in range(self.times):
            press("tab", interval=self.pause)


class Button(Control):
    def __init__(self, pause=0):
        self.pause = pause

    def fill(self):
        if self.verbose:
            print("Click button")
        press("enter", interval=self.pause)


class Pause(Control):
    """Pause for a moment"""

    def __init__(self, pause):
        self.pause = pause

    def fill(self):
        if self.verbose:
            print(f"Pause {self.pause} seconds")
        time.sleep(self.pause)


class TextField(Control):
    """Text field"""

    def __init__(self, data, pause=0):
        """data: text to fill"""
        self.data = data
        self.pause = pause

    def fill(self):
        if self.verbose:
            print(f"Fill text field with {self.data}")
        # press(self.data.replace(" ", "{SPACE}"))
        write(self.data, interval=self.pause)
        press("tab")


class RadioButton(Control):
    """Radio button"""

    def __init__(self, data, pause=0):
        """data: False for No, True for yes"""
        self.data = data
        self.pause = pause

    def fill(self):
        if self.data:
            key = "right"  # TODO:
        else:
            key = "space"
        if self.verbose:
            print(f"Fill radio button with {self.data}, press {key} key")
        press(key, interval=self.pause)
        press("tab", interval=self.pause)


class CheckBox(Control):
    """Check box"""

    def __init__(self, data, pause=0):
        self.data = data
        self.pause = pause

    def fill(self):
        if self.verbose:
            print(f"Fill check box with {self.data}")
        if self.data:
            press("space", interval=self.pause)
            press("tab", interval=self.pause)
        else:
            press("tab", interval=self.pause)


class OutputInfo(Control):
    """Output log info"""

    def __init__(self, info, pause=0):
        self.info = info
        self.pause = pause

    def fill(self):
        if self.verbose:
            print(f"------- {self.info} -------")
            time.sleep(self.pause)


class DropdownList(Control):
    """Dropdown list"""

    def __init__(self, data, options, like: bool = False, pause=0):
        self.data = data
        self.options = options
        self.like = like
        self.pause = pause

    def key_presses(self) -> Tuple[str, int]:
        s_char = self.data[0].lower()
        count = 0
        matched_one = best_match(self.data, self.options)
        for elem in self.options:
            if elem[0].lower() == s_char:
                count += 1
                if elem.lower() == self.data.lower():
                    break
                elif self.like and f"({self.data})" in elem:
                    break
                elif elem.lower() == matched_one.lower():
                    break
        return (s_char, count)

    def fill(self):
        key, num = self.key_presses()
        if self.verbose:
            print(f"Select dropdown list with {self.data}, press {key} {num} time(s)")
        press("down", interval=self.pause)
        press(key, presses=num, interval=0.1)
        press("enter")
        press("tab", interval=self.pause)


class DateField(Control):
    """Date field"""

    def __init__(self, date, noday: bool = False, pause=0.2):
        self.date = date
        self.noday = noday
        self.pause = pause

    def fill(self):
        if self.noday:
            year, month = self.date.split("-")
            if self.verbose:
                print(f"Fill date {year}-{month}")
            write(year, interval=self.pause)
            write(month, interval=self.pause)
        else:
            year, month, day = self.date.split("-")
            if self.verbose:
                print(f"Fill date {year}-{month}-{day}")
            write(year, interval=self.pause)
            write(month, interval=self.pause)
            write(day, interval=self.pause)
        press("tab", interval=self.pause)


# TODO: Test if press needs interval?



from pydantic import BaseModel, validator
from typing import Optional
from datetime import date, timedelta
from pydantic.class_validators import root_validator
from utils.utils import speaking_list

class JobofferBase(BaseModel):
    job_title: str
    noc: str
    hours: float
    days: int
    wage_unit: str
    wage_rate: float
    ot_ratio: float
    permanent: bool
    work_start_date: Optional[date]
    job_duration: Optional[float]
    job_duration_unit: Optional[str]
    disability_insurance: bool
    dental_insurance: bool
    empolyer_provided_persion: bool
    extended_medical_insurance: bool
    extra_benefits: Optional[str]
    
    @property
    def has_benefits(self):
        return any([self.disability_insurance,self.dental_insurance,self.empolyer_provided_persion,self.extended_medical_insurance,self.extra_benefits])

    @property
    def benefits(self):
        benefits_list=[]
        if  self.disability_insurance:benefits_list.append("Disability insurance")
        if  self.dental_insurance:benefits_list.append("Dental insurance")
        if  self.empolyer_provided_persion:benefits_list.append("Empolyer provided persion")
        if  self.extended_medical_insurance:benefits_list.append("Extended medical insurance")
        if  self.extra_benefits:benefits_list.append(self.extra_benefits)
        return speaking_list(benefits_list)
    
    @root_validator
    def checkDuration(cls, values):
        if not values.get("permanent") and (
            not values.get("job_duration") or not values.get("job_duration_unit")
        ):
            raise ValueError(
                "Since it is not permanent job offer, so you have to specify the job duration and job duration unit in info-joboffer sheet"
            )
        return values

    @property
    def term(self):
        if self.permanent:
            return "Permanent"
        else:
            if not self.job_duration or not self.job_duration_unit:
                raise ValueError(
                    "Since you claimed the job offer is not permanent, you must spcify the job duration and duration unit"
                )
            return str(self.job_duration) + " " + self.job_duration_unit

    @property
    def full_part_time(self):
        return "full-time" if self.hours >= 30 else "part-time"

    @property
    def is_full_time(self):
        return True if self.hours >= 30 else False

    @property
    def salary(self):
        return "{:,.1f}".format(self.wage_rate)

    @property
    def weekly_hours(self):
        return "{:,.1f}".format(self.hours)

    @property
    def hourly_rate(self):
        rate = 0
        match self.wage_unit:
            case "annually":
                rate = self.wage_rate / 52 / self.hours
            case "monthly":
                rate = self.wage_rate * 12 / 52 / self.hours
            case "weekly":
                rate = self.wage_rate / self.hours
            case "hourly":
                rate = self.wage_rate
        return "{0:.4g}".format(rate)

    @property
    def how_to_convert_to_hourly_rate(self):
        the_way = ""
        match self.wage_unit:
            case "annually":
                rate = self.wage_rate / 52 / self.hours
                the_way = f"The annual wage is {'{0:.10g}'.format(self.wage_rate)}, and weekly working hours is {'{0:.4g}'.format(self.hours)}, so the hourly rate is {self.wage_rate}/52/{self.hours}={'{0:.4g}'.format(rate)}"
            case "monthly":
                rate = self.wage_rate * 12 / 52 / self.hours
                the_way = f"The monthly wage is {'{0:.10g}'.format(self.wage_rate)}, and weekly working hours is {'{0:.4g}'.format(self.hours)}, so the hourly rate is {self.wage_rate}*12/52/{self.hours}={'{0:.4g}'.format(rate)}"
            case "weekly":
                rate = self.wage_rate / self.hours
                the_way = f"The weekly wage is {'{0:.10g}'.format(self.wage_rate)}, and weekly working hours is {'{0:.4g}'.format(self.hours)}, so the hourly rate is {self.wage_rate}/{self.hours}={'{0:.4g}'.format(rate)}"
            case "hourly":
                rate = self.wage_rate
        return the_way

    @property
    def is_hourly_rate_converted(self):
        return False if self.wage_unit == "hourly" else True

    @property
    def hourly_rate_say(self):
        return f"${self.hourly_rate} per hour"

    @property
    def overtime_rate(self):
        return "{0:.4g}".format(float(self.hourly_rate) * self.ot_ratio)

    @property
    def overtime_rate_say(self):
        return "${rate:,.0f} per hour".format(rate=float(self.overtime_rate))

    @property
    def weekly_rate(self):
        rate = 0
        match self.wage_unit:
            case "annually":
                rate = self.wage_rate / 52
            case "monthly":
                rate = self.wage_rate * 12 / 52
            case "weekly":
                rate = self.wage_rate
            case "hourly":
                rate = self.wage_rate * self.hours
        return "{0:.0f}".format(rate)

    @property
    def weekly_rate_say(self):
        return "${rate:,.0f} per week".format(rate=float(self.weekly_rate))

    @property
    def monthly_rate(self):
        rate = 0
        match self.wage_unit:
            case "annually":
                rate = self.wage_rate / 12
            case "monthly":
                rate = self.wage_rate
            case "weekly":
                rate = self.wage_rate * 52 / 12
            case "hourly":
                rate = self.wage_rate * self.hours * 52 / 12
        return "{0:.0f}".format(rate)

    @property
    def monthly_rate_say(self):
        return "${rate:,.0f} per month".format(rate=float(self.monthly_rate))

    @property
    def annual_rate(self):
        rate = 0
        match self.wage_unit:
            case "annually":
                rate = self.wage_rate
            case "monthly":
                rate = self.wage_rate * 12
            case "weekly":
                rate = self.wage_rate * 52
            case "hourly":
                rate = self.wage_rate * self.hours * 52
        return "{0:.0f}".format(rate)

    @property
    def annual_rate_say(self):
        return "${rate:,.0f} per year".format(rate=float(self.annual_rate))

    @property
    def requirements(self):
        reqs = [
            self.specific_edu_requirement,
            self.skill_experience_requirement,
            *self.other_requirements,
        ]
        return [req for req in reqs if req]

    # work start date + days (base on job_duration and job_duration_unit)
    @property
    def work_end_date(self):
        days = 0

        match self.job_duration_unit:
            case "day":
                return self.work_start_date + timedelta(days=int(self.job_duration))
            case "week":
                return self.work_start_date + timedelta(weeks=int(self.job_duration))
            case "month":
                return self.work_start_date + timedelta(
                    days=int(self.job_duration * 30)
                )
            case "year":
                return self.work_start_date + timedelta(
                    days=int(self.job_duration * 365)
                )

    @property
    def is_primary_agriculture_position(self):
        return self.noc in ["0821", "0822", "8252", "8255", "8431", "8432", "8611"]

    @property
    def lmia_advertisement_variation(self):
        """Employers who wish to hire foreign workers in the following categories are subject to a variation in the advertising requirements:
        https://www.canada.ca/en/employment-social-development/services/foreign-workers/variations.html"""
        return self.noc in []  # TODO:



from .context import DATADIR
from pydantic import BaseModel
from model.bcpnp.data import General
from model.common.jobposition import PositionBase
from model.common.wordmaker import WordMaker
import os
from model.common.commonmodel import CommonModel


class Position(PositionBase):
    pass


class CompanyInfoModel(BaseModel):
    general: General
    position: Position


class CompanyInfoModel_E(CommonModel, CompanyInfoModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = ["excel/er.xlsx"]
        super().__init__(excels, output_excel_file, mother_excels, globals())


class CompanyInfoDocxAdaptor:
    def __init__(self, employer_training_obj: CompanyInfoModel):
        self.employer_training_obj = employer_training_obj

    def make(self, output_docx):
        template_path = os.path.abspath(
            os.path.join(DATADIR, "word/bcpnp_company_information.docx")
        )
        wm = WordMaker(template_path, self.employer_training_obj, output_docx)
        wm.make()



from model.lmia.common import *
from model.common.commonmodel import CommonModel
from webform.lmiaportal.webformmodel import WebformModel
from model.common.contact import Contacts
from model.common.finance import Finance
from model.common.advertisement import (
    InterviewRecord,
    InterviewRecords,
    RecruitmentSummary,
)
import os


class Emp5627(BaseModel):
    named: bool
    provide_accommodation: bool
    description: Optional[str]
    rent_unit: str
    rent_amount: float
    accommodation_type: str
    explain: Optional[str]
    bedrooms: str
    people: str
    bathrooms: str
    other: Optional[str]
    cap_exempted: bool
    which_exemption: Optional[str]
    exemption_rationale: Optional[str]
    is_in_seasonal_industry: Optional[bool]
    four_week_start_date: Optional[date]
    four_week_end_date: Optional[date]
    q_a: Optional[int]
    q_b: Optional[int]
    q_c: Optional[int]
    q_d: Optional[int]
    q_e: Optional[int]
    q_f: Optional[int]
    q_g: Optional[int]
    q_h: Optional[int]

    @root_validator
    def checkAnswers(cls, values):
        cap_exempted = values.get("cap_exempted", None)
        explanations = [
            values.get(var, None)
            for var in [
                "which_exemption",
                "exemption_rationale",
            ]
        ]
        if cap_exempted and not all(v is not None for v in explanations):
            raise ValueError(
                "Since CAP exempted, you didn't input all the information about the which exemptin and/or the rationale."
            )

        not_exempted_info = [
            values.get(var, None)
            for var in [
                "is_in_seasonal_industry",
                "four_week_start_date",
                "four_week_end_date",
                "q_a",
                "q_b",
                "q_c",
                "q_d",
                "q_e",
                "q_f",
                "q_g",
                "q_h",
            ]
        ]
        if not cap_exempted and not all(v is not None for v in not_exempted_info):
            raise ValueError(
                "Since the application is not CAP exempted, you didn't input all the information required for CAP claculation."
            )

        provide_accommodation = values.get("provide_accommodation", None)
        description = values.get("description", None)
        if not provide_accommodation and not description:
            raise ValueError(
                "You don't provide accommodation, but did not provide the explaination on how to assist the TFW."
            )

        provide_accommodation = values.get("provide_accommodation", None)
        provide_accommodation_details = [
            values.get(var, None)
            for var in [
                "rent_unit",
                "rent_amount",
                "accommodation_type",
                "bedrooms",
                "people",
                "bathrooms",
            ]
        ]
        if provide_accommodation and not all(
            v is not None for v in provide_accommodation_details
        ):
            raise ValueError(
                "Since employer provides accommodation, you didn't input all the information about the accommodation."
            )

        accommodation_type = values.get("accommodation_type", None)
        explain = values.get("explain", None)
        if accommodation_type == "other" and not explain:
            raise ValueError(
                "Since accommodation type is 'other', but you didn't describe it."
            )
        return values


class Personal(Person):
    last_name: Optional[str]
    first_name: Optional[str]
    sex: Optional[str]
    dob: Optional[date]
    citizen: Optional[str]


class M5627Model(BaseModel):
    lmiacase: LmiaCase
    general: General
    lmi: Lmi
    emp5627: Emp5627
    eraddress: List[ErAddress]
    contact: List[Contact]
    finance: List[Finance]
    joboffer: Joboffer
    personal: Personal
    personalassess: PersonalAssess
    position: Position
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    rcic: Rcic
    recruitmentsummary: RecruitmentSummary

    def checkTFW(self):
        if self.emp5627.named and not all(
            [
                self.personal.first_name,
                self.personal.last_name,
                self.personal.sex,
                self.personal.citizen,
                self.personal.dob,
            ]
        ):
            raise ValueError(
                "Since TFW is name required, you haven't input all information required in info-personal sheet."
            )

    def make_web_form(self, output_json, upload_dir, rcic, *args, **kwargs):
        args = {
            "model_variable": "5627",
            "app": self,
            "output_json": output_json,
            "upload_dir": upload_dir,
            "rcic": rcic,
        }
        wf = WebformModel(**args)
        wf.save()

    def make_pdf_form(self, *args, **kwargs):
        pass

    def context(self, *args, **kwargs):
        interview = InterviewRecords(self.interviewrecord)
        primary_contact = Contacts(self.contact).primary
        return {
            **self.dict(),
            "formatted_finance": [
                {
                    "year": f.year,
                    "formatted_revenue": f.formatted_revenue,
                    "formatted_net_income": f.formatted_net_income,
                    "formatted_retained_earning": f.formatted_retained_earning,
                }
                for f in self.finance
            ],
            "num_of_job_posts": Advertisements(self.advertisement).amount,
            "why_qualified": self.personalassess.why_qualified,
            "adv_summary": [
                {
                    "start_date": a.start_date,
                    "end_date": a.end_date,
                    "media": a.media,
                    "days": a.days,
                }
                for a in self.advertisement
            ],
            "high_wage": True
            if float(self.joboffer.hourly_rate) >= self.lmiacase.provincial_median_wage
            else False,
            "rs": {
                "resume_num": interview.resume_num,
                "canadian_num": interview.canadian_num,
                "unknown_num": interview.unknown_num,
                "foreigner_num": interview.foreigner_num,
                "total_canadian": interview.total_canadian,
                "total_interviewed_canadians": interview.total_interviewed_canadians,
                "canadian_records": interview.canadian_records,
            },
            "primary_contact": primary_contact,
            "within_2years": self.general.within_2years,
            "lmi_benefits": [
                self.lmi.job_creation_benefit,
                self.lmi.skill_transfer_benefit,
                self.lmi.fill_shortage_benefit,
                self.lmi.other_benefit,
            ],
            "duration": self.lmiacase.duration,
            "why_qualified": [
                item
                for item in [
                    self.personalassess.work_experience_brief,
                    self.personalassess.education_brief,
                    self.personalassess.competency_brief,
                    self.personalassess.language_brief,
                    self.personalassess.performance_remark,
                ]
                if item
            ],
            "employee_name": self.personal.full_name,
        }


class M5627ModelE(CommonModel, M5627Model):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/recruitment.xlsx",
            "excel/pa.xlsx",
            "excel/er.xlsx",
            "excel/rep.xlsx",
            "excel/lmia.xlsx",
        ]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from pydantic import BaseModel, validator, EmailStr
from datetime import date
from typing import Optional, List
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.address import Address, Addresses
from model.common.utils import makeList
from pydantic.class_validators import root_validator


class ErAddress(Address):
    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()


class ErAddresses(Addresses):
    def __init__(self, address_list: List[Address]) -> None:
        super().__init__(address_list)


class General(EmployerBase):
    company_intro: str
    business_intro: str
    recruit_email: EmailStr


class JobOffer(JobofferBase):
    offer_date: date
    supervisor_name: Optional[str]
    supervisor_title: Optional[str]
    vacation_pay_days: int
    vacation_pay_percentage: float
    employer_rep: str
    employer_rep_title: str
    payment_way: str
    has_probation: bool
    probation_duration: Optional[int]
    duties_brief: str
    duties: list

    _str2bool_duties = validator("duties", allow_reuse=True, pre=True)(makeList)

    @root_validator
    def checkProbation(cls, values):
        if values.get("has_probation") and not values.get("probation_duration"):
            raise ValueError(
                "Since it is has probation period, but you did not specify the probation duration in info-joboffer sheet"
            )
        return values

    @property
    def date_of_offer(self):
        return self.offer_date.strftime("%b %d, %Y")

    @property
    def vacation_pay_percent(self):
        return "{:,.1f}%".format(self.vacation_pay_percentage * 100)



from .context import DATADIR
from typing import List
from model.common.commonmodel import CommonModel, BuilderModel
from model.bcpnp.data import Bcpnp, General, JobOffer, ErAddress
from model.common.jobposition import PositionBase
from model.common.rcic import Rcic
from model.common.advertisement import (
    Advertisement,
    Advertisements,
    InterviewRecord,
    InterviewRecords,
    RecruitmentSummary,
)
from model.common.person import Person, PersonalAssess
from model.common.address import Addresses
from model.common.wordmaker import WordMaker
import os
from pydantic import BaseModel


class Personal(Person):
    def __str__(self):
        return self.full_name


class Position(PositionBase):
    pass


class EmployerTrainingModel(BaseModel, BuilderModel):
    eraddress: List[ErAddress]
    general: General
    position: Position
    personal: Personal
    joboffer: JobOffer
    personalassess: PersonalAssess
    bcpnp: Bcpnp
    rcic: Rcic
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    recruitmentsummary: RecruitmentSummary

    @property
    def work_location(self):
        addresses = Addresses(self.eraddress)
        return addresses.working

    @property
    def selected_contact(self):
        contacts = Contacts(self.contact)
        return contacts.preferredContact

    @property
    def summary(self):
        return InterviewRecords(self.interviewrecord)

    @property
    def advertisements(self):
        return Advertisements(self.advertisement)

    @property
    def person(self):
        return {
            "full_name": self.personal.full_name,
            "attributive": self.personal.attributive,
            "object": self.personal.object,
            "subject": self.personal.subject,
            "short_name": self.personal.short_name,
        }

    def context(self, *args, **kwargs):
        context = {
            **self.dict(),
            "resume_num": self.summary.resume_num,
            "canadian_num": self.summary.canadian_num,
            "unknown_num": self.summary.unknown_num,
            "foreigner_num": self.summary.foreigner_num,
            "total_canadian": self.summary.total_canadian,
            "total_interviewed_canadians": self.summary.total_interviewed_canadians,
            "canadian_records": self.summary.canadian_records,
            "advertisement": self.advertisements,
            "personal": self.person,
            "date_of_offer": self.joboffer.date_of_offer,
            "work_start_date": self.joboffer.start_date_say,
            "joboffer_date": self.joboffer.date_of_offer,
            "work_location": self.work_location,
        }
        return context

    def make_pdf_form(self, *args, **kwargs):
        pass

    def make_web_form(self, *args, **kwargs):
        pass


class EmployerTrainingModelE(CommonModel, EmployerTrainingModel):
    def __init__(self, excels=None, output_excel_file=None):
        mother_excels = [
            "excel/er.xlsx",
            "excel/pa.xlsx",
            "excel/recruitment.xlsx",
            "excel/bcpnp.xlsx",
            "excel/rep.xlsx",
        ]
        super().__init__(excels, output_excel_file, mother_excels, globals())



from webform.definition import Action
from webform.webcommon import WebPages


class Dashboard(WebPages):
    @property
    def actions(self):
        return [self.pickEmployer, self.createApplication]

    @property
    def createApplication(self):
        create_button_id = "body > main > ul > li:nth-child(2) > a"
        next_page_tag = "#\\39 124 > select"
        return self.web_element.pageElement(
            create_button_id,
            next_page_tag,
            [],
            label="Dashboard create LMIA Application",
        )

    @property
    def pickEmployer(self):
        next_page_tag = "#wb-auto-11_filter > label > input[type=search]"  # filter
        action = {
            "action_type": Action.LmiaEmployerPick.value,
            "business_number": self.app.general.cra_number,
        }
        return self.web_element.pageElement(
            None, next_page_tag, actions=[action], label="Dashboard pick employer"
        )



from pydantic import BaseModel, EmailStr, validator
from typing import Optional
from model.common.utils import normalize, trimString
from datetime import date, datetime
from dateutil.parser import parse, ParserError


class Person(BaseModel):
    last_name: str
    first_name: str
    sex: str
    dob: date
    citizen: str

    _normalize_first_name = validator(
        "first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_last_name = validator("last_name", allow_reuse=True, check_fields=False)(
        normalize
    )

    @property
    def full_name(self):
        return self.first_name + " " + self.last_name

    @property
    def age(self):
        the_day = date.today()
        try:
            return (
                the_day.year
                - self.dob.year
                - ((the_day.month, the_day.day) < (self.dob.month, self.dob.day))
            )
        except (ValueError, TypeError, ParserError) as err:
            print(err.args[0], " in file ", __file__)

    def ageOnDate(self, the_day=date.today()):
        if not isinstance(the_day, date):
            try:
                the_day = datetime.strptime(the_day, "%Y-%m-%d")
                return (
                    the_day.year
                    - self.dob.year
                    - ((the_day.month, the_day.day) < (self.dob.month, self.dob.day))
                )
            except (ValueError, TypeError) as err:
                print(err.args[0], " in file ", __file__)

    # Third person subject
    @property
    def subject(self):
        return "he" if self.sex.upper() == "M" or self.sex.upper() == "MALE" else "she"

    # Third person object
    @property
    def object(self):
        return "him" if self.sex.upper() == "M" or self.sex.upper() == "MALE" else "her"

    # third person attributive
    @property
    def attributive(self):
        return "his" if self.sex.upper() == "M" or self.sex.upper() == "MALE" else "her"

    # salutation
    @property
    def salutation(self):
        return "Mr." if self.sex.upper() == "M" or self.sex.upper() == "MALE" else "Ms."

    @property
    def short_name(self):
        return self.salutation + " " + self.last_name

    @property
    def respectful_full_name(self):
        return self.salutation + " " + self.full_name

    @property
    def birthday(self):
        return self.dob.strftime("%b %d, %Y")


class PersonalAssess(BaseModel):
    work_experience_brief: str
    education_brief: str
    competency_brief: str
    language_brief: Optional[str]
    performance_remark: Optional[str]

    @property
    def why_qualified(self):
        qualifications = [
            self.work_experience_brief,
            self.education_brief,
            self.competency_brief,
            self.language_brief,
            self.performance_remark,
        ]
        return [q for q in qualifications if q is not None]

    @property
    def why_qualified_say(self):
        return " ".join(self.why_qualified)


def main():
    p1 = {
        "first_name": "jacky",
        "last_name": "zhang",
        "dob": "1969-02-11",
        "sex": "Male",
        "citizen": "China",
    }
    jacky = Person(**p1)
    print(jacky.full_name, jacky.age, jacky.ageOnDate("2056-02-11"))


if __name__ == "__main__":
    main()
